(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory CPF_Parser
imports
  Proof_Checker
  Xmlt
begin

fun var :: "('a, string) term xmlt" where "var x = (Xmlt.text ''var'' x \<guillemotright>= (return \<circ> Var))"
hide_const (open) var

partial_function (sum_bot)
  "term" :: "'a xmlt \<Rightarrow> ('a, string) term xmlt"
where
  [code]: "term xml2name x = Xmlt.options [
    (''var'', CPF_Parser.var),
    (''funapp'', Xmlt.many1 ''funapp'' xml2name (
      Xmlt.singleton ''arg'' (term xml2name) id) Fun)
  ] x"  
hide_const (open) "term"

partial_function (sum_bot) ctxt :: "'a xmlt \<Rightarrow> ('a, string) ctxt xmlt" where
  [code]: "ctxt xml2name x = Xmlt.options [
     (''box'', Xmlt.leaf ''box'' Hole),  
     (''funContext'', Xmlt.tuple4 ''funContext'' 
       xml2name 
       (Xmlt.many ''before'' (CPF_Parser.term xml2name) id)
       (ctxt xml2name)
       (Xmlt.many ''after'' (CPF_Parser.term xml2name) id)
       More)
   ] x"  

definition flat_contexts :: "'a xmlt \<Rightarrow> ('a, string) ctxt list xmlt" where
  "flat_contexts xml2name = Xmlt.many ''flatContexts'' (CPF_Parser.ctxt xml2name) id"
hide_const (open) flat_contexts

definition subst :: "'a xmlt \<Rightarrow> ('a, string) substL xmlt" where
  "subst xml2name = Xmlt.many ''substitution''
    (Xmlt.pair ''substEntry'' 
      (Xmlt.text ''var'')
      (CPF_Parser.term xml2name)
      Pair
    ) id"
hide_const (open) subst

primrec list2position :: "nat list \<Rightarrow> pos" where
  "list2position [] = \<epsilon>"
| "list2position (n # ns) = n <# list2position ns"

definition position :: "nat xmlt" where
  "position \<equiv> Xmlt.change (Xmlt.nat ''position'') (\<lambda>n. n - 1)"
hide_const (open) position

definition pos :: "pos xmlt" where
  "pos \<equiv> Xmlt.many ''positionInTerm'' CPF_Parser.position list2position"
hide_const (open) pos


definition afs :: "'a xmlt \<Rightarrow> 'a afs_list xmlt" where
  "afs xml2name \<equiv> Xmlt.many ''argumentFilter'' (Xmlt.triple ''argumentFilterEntry''
    xml2name
    (Xmlt.nat ''arity'')
    (Xmlt.options [
      (''collapsing'', Xmlt.change (Xmlt.nat ''collapsing'') (\<lambda>i. Collapse (i - 1))),
      (''nonCollapsing'', Xmlt.many ''nonCollapsing'' CPF_Parser.position AFList)
    ])
    (\<lambda>f a e. ((f, a), e))
  ) id"
hide_const (open) afs

definition proj :: "'a xmlt \<Rightarrow> 'a projL xmlt" where
  "proj xml2name \<equiv> Xmlt.change (CPF_Parser.afs xml2name) (\<lambda>afl. Projection 
    (map (\<lambda>(fa, e). case e of 
      Collapse i \<Rightarrow> Pair fa i
    | _ \<Rightarrow> Pair fa (snd fa)) afl))" 

definition rule :: "'a xmlt \<Rightarrow> ('a, string) rule xmlt" where
  "rule xml2name = Xmlt.pair ''rule''
    (Xmlt.singleton ''lhs'' (CPF_Parser.term xml2name) id)
    (Xmlt.singleton ''rhs'' (CPF_Parser.term xml2name) id)
    Pair"
hide_const (open) rule 

definition
  relstep :: "'a xmlt \<Rightarrow> (pos \<times> ('a, string) rule \<times> bool \<times> ('a, string) term) xmlt"
where
  "relstep xml2name = xml3to4elements ''rewriteStep''
    CPF_Parser.pos
    (CPF_Parser.rule xml2name)
    (Xmlt.leaf ''relative'' ())
    (CPF_Parser.term xml2name)
    (\<lambda>p r rel t. (p, r, rel = None, t))"
hide_const (open) relstep

definition
  rstep :: "'a xmlt \<Rightarrow> (pos \<times> ('a, string) rule \<times> ('a, string) term) xmlt"
where
  "rstep xml2name = Xmlt.triple ''rewriteStep''
    CPF_Parser.pos
    (CPF_Parser.rule xml2name)
    (CPF_Parser.term xml2name)
    (\<lambda>p r t. (p, r, t))"
hide_const (open) rstep

definition
  estep :: "'a xmlt \<Rightarrow> (pos \<times> ('a, string) rule \<times> bool \<times> ('a, string) term) xmlt"
where
  "estep xml2name = Xmlt.tuple4 ''equationStep''
    CPF_Parser.pos
    (CPF_Parser.rule xml2name)
    (Xmlt.options [
       (''leftRight'', Xmlt.leaf ''leftRight'' True),
       (''rightLeft'', Xmlt.leaf ''rightLeft'' False)
    ])
    (CPF_Parser.term xml2name)
    (\<lambda>p r b t. (p, r, b, t))"
hide_const (open) estep

definition
  relsteps :: "'a xmlt \<Rightarrow> (('a,string)term \<times> ('a,string)prseq) xmlt"
where
  "relsteps xml2name \<equiv> Xmlt.many1 ''rewriteSequence'' 
    (Xmlt.singleton ''startTerm'' (CPF_Parser.term xml2name) id)
    (CPF_Parser.relstep xml2name)
    Pair"
hide_const (open) relsteps

definition
  rsteps :: "'a xmlt \<Rightarrow> (('a,string)term \<times> ('a,string)rseq) xmlt"
where
  "rsteps xml2name \<equiv> Xmlt.many1 ''rewriteSequence'' 
    (Xmlt.singleton ''startTerm'' (CPF_Parser.term xml2name) id)
    (CPF_Parser.rstep xml2name)
    Pair"
hide_const (open) rsteps


definition pat_eqv_prf :: "'a xmlt \<Rightarrow> ('a,string)pat_eqv_prf xmlt"
  where "pat_eqv_prf xml2name \<equiv> let sub = CPF_Parser.subst xml2name in
   Xmlt.singleton ''patternEquivalence''
    (Xmlt.options [
       (''domainRenaming'', Xmlt.singleton ''domainRenaming'' sub Pat_Dom_Renaming),
       (''irrelevant'', Xmlt.pair ''irrelevant''  sub sub Pat_Irrelevant),
       (''simplification'', Xmlt.pair ''simplification'' sub sub Pat_Simplify)
     ]) id"
hide_const (open) pat_eqv_prf

definition pat_term :: "'f xmlt \<Rightarrow> (('f,string)term \<times> ('f,string)substL \<times> ('f,string)substL) xmlt"
  where "pat_term xml2name \<equiv> Xmlt.triple ''patternTerm''
            (CPF_Parser.term xml2name)
            (CPF_Parser.subst xml2name)
            (CPF_Parser.subst xml2name)
            (\<lambda> t s1 s2. (t,s1,s2))"


abbreviation (input) pat_rule_fun where "pat_rule_fun xml2name x pat \<equiv> (let 
   sub = CPF_Parser.subst xml2name;
   pt = pat_term xml2name;
   var = Xmlt.text ''var''
 in 
   Xmlt.triple ''patternRule'' pt pt
    (Xmlt.options [
       (''originalRule'', Xmlt.pair ''originalRule'' (CPF_Parser.rule xml2name) (Xmlt.bool ''isPair'')  Pat_OrigRule),
       (''initialPumping'', Xmlt.triple ''initialPumping'' pat sub sub Pat_InitPump),
       (''initialPumpingContext'', Xmlt.tuple4 ''initialPumpingContext'' pat sub CPF_Parser.pos var Pat_InitPumpCtxt),
       (''equivalence'', Xmlt.triple ''equivalence'' pat (Xmlt.options [
                  (''left'',Xmlt.leaf ''left'' True),
                  (''right'',Xmlt.leaf ''right'' False)
             ]) (CPF_Parser.pat_eqv_prf xml2name) Pat_Equiv),
       (''narrowing'', Xmlt.triple ''narrowing'' pat pat CPF_Parser.pos Pat_Narrow),
       (''instantiation'', Xmlt.triple ''instantiation'' pat sub (Xmlt.options [
                  (''base'',Xmlt.leaf ''base'' Pat_Base),
                  (''pumping'',Xmlt.leaf ''pumping'' Pat_Pump),
                  (''closing'',Xmlt.leaf ''closing'' Pat_Close)
             ]) Pat_Inst),
       (''instantiationPumping'', Xmlt.pair ''instantiationPumping'' pat 
             (Xmlt.nat ''power'')
             Pat_Exp_Sigma),
       (''rewriting'', Xmlt.triple ''rewriting'' pat (CPF_Parser.rsteps xml2name)  (Xmlt.options [
                  (''base'',Xmlt.leaf ''base'' (Pat_Base,[])),
                  (''pumping'',Xmlt.singleton ''pumping'' var (Pair Pat_Pump)),
                  (''closing'',Xmlt.singleton ''closing'' var (Pair Pat_Close))
             ]) (\<lambda> p seq (po,va). Pat_Rewr p seq po va))
     ]) (\<lambda> _ _ z. z) x)" 

partial_function (sum_bot) pat_rule_prf :: "'a xmlt \<Rightarrow> ('a,string)pat_rule_prf xmlt"
  where "pat_rule_prf xml2name x = pat_rule_fun xml2name x (pat_rule_prf xml2name)"

(* one cannot put the 'let pat = ... ' directly into partial function, since then the monotonicity fails *)
lemma pat_rule_prf [code]: 
  "pat_rule_prf xml2name x = (let pat = pat_rule_prf xml2name in pat_rule_fun xml2name x pat)"
     unfolding pat_rule_prf.simps[of _ x] Let_def by simp

hide_const (open) pat_rule_prf


definition
  conversion :: "'a xmlt \<Rightarrow> (('a,string)eseq) xmlt"
where
  "conversion xml2name \<equiv> Xmlt.many1 ''conversion'' 
    (Xmlt.singleton ''startTerm'' (CPF_Parser.term xml2name) id)
    (CPF_Parser.estep xml2name)
    (\<lambda> _ x. x)"
hide_const (open) conversion

definition subsumption_proof :: "'a xmlt \<Rightarrow> ('a,string)subsumption_proof xmlt"
  where "subsumption_proof xml2name \<equiv> Xmlt.many ''subsumptionProof''
     (Xmlt.pair ''ruleSubsumptionProof''
        (CPF_Parser.rule xml2name)
        (CPF_Parser.conversion xml2name)
        Pair
    ) id"
hide_const (open) subsumption_proof

definition joinable_critical_pairs :: "'a xmlt \<Rightarrow> ('a,string)crit_pair_joins xmlt"
  where "joinable_critical_pairs xml2name \<equiv> let rew = CPF_Parser.rsteps xml2name in
    Xmlt.many ''joinableCriticalPairs'' 
      (Xmlt.pair ''joinableCriticalPair'' rew rew (\<lambda> (s,sseq) (t,tseq). (s,sseq,t,tseq))) id"

definition wcr_proof :: "'a xmlt \<Rightarrow> 'a join_info xmlt"
  where "wcr_proof xml2name \<equiv> Xmlt.singleton ''wcrProof'' 
    (Xmlt.options [
       (''joinableCriticalPairs'', Xmlt.change (joinable_critical_pairs xml2name) Guided),
       (''joinableCriticalPairsBFS'', Xmlt.change (Xmlt.nat ''joinableCriticalPairsBFS'') Join_BFS),
       (''joinableCriticalPairsAuto'', Xmlt.leaf ''joinableCriticalPairsAuto'' Join_NF)
  ]) id"

definition
  forbidden_pattern :: "'a xmlt \<Rightarrow> ((('a, string) ctxt \<times> ('a, string) term \<times> Forbidden_Patterns.location) xmlt)"
where
  "forbidden_pattern xml2name xml \<equiv> ((Xmlt.triple ''forbiddenPattern''
    (CPF_Parser.term xml2name)
    CPF_Parser.pos
    (Xmlt.options [
       (''here'', Xmlt.leaf ''here'' Forbidden_Patterns.H),
       (''above'', Xmlt.leaf ''above'' Forbidden_Patterns.A),
       (''below'', Xmlt.leaf ''below'' Forbidden_Patterns.B)
    ]) 
    (\<lambda> t p l. (t,p,l)) xml) \<guillemotright>= 
   (\<lambda> (t,p,l).
   if p \<in> poss t then return (ctxt_of_pos_term p t, t |_ p, l) else error (''position does not exist in term'')))"
hide_const (open) forbidden_pattern

definition
  forbidden_patterns :: "'a xmlt \<Rightarrow> ((('a, string) ctxt \<times> ('a, string) term \<times> Forbidden_Patterns.location) list) xmlt"
where
  "forbidden_patterns xml2name \<equiv> Xmlt.many ''forbiddenPatterns''
    (CPF_Parser.forbidden_pattern xml2name)
    id"
hide_const (open) forbidden_patterns

definition 
  loop ::
    "'a xmlt \<Rightarrow>
    (('a, string) term \<times> ('a, string) prseq \<times> ('a, string) substL \<times> ('a, string) ctxt) xmlt"
where
  "loop xml2name \<equiv> Xmlt.triple ''loop''
    (CPF_Parser.relsteps xml2name)
    (CPF_Parser.subst xml2name)
    (CPF_Parser.ctxt xml2name)
    (\<lambda>(s, rseq) \<sigma> c. (s, rseq, \<sigma>, c))"
hide_const (open) loop

definition
  start_term :: "'a xmlt \<Rightarrow> ('a, string) term \<Rightarrow> ('a, string) term xmlt"
where
  "start_term xml2name t = Xmlt.singleton ''startTerm'' (\<lambda>xml. (
     CPF_Parser.term xml2name xml \<guillemotright>= (\<lambda>s. if s = t
       then return t
       else error (''<startTerm> does not match lhs'')
     )
   )) id"
hide_const (open) start_term

definition
  rseq ::
    "'a xmlt \<Rightarrow> 'a proj \<Rightarrow> ('a, string) rule \<Rightarrow> (('a,string)rule \<times> ('a,string)rseq)xmlt"
where
  "rseq xml2name \<pi> r = Xmlt.many1 ''rewriteSequence''
     (CPF_Parser.start_term xml2name (proj_term \<pi> (fst r)))
     (CPF_Parser.rstep xml2name)
     (\<lambda>_ rseq. (r, rseq))"
hide_const (open) rseq

fun
  projected_rseq ::
    "'a xmlt \<Rightarrow> 'a proj \<Rightarrow>
    (('a, string) rule \<times> ('a,string)rseq) xmlt"
where
  "projected_rseq xml2name \<pi> (XML name atts cs text) = (
     let tag = ''projectedRewriteSequence'' in (
     if name = tag \<and> atts = [] \<and> text = None \<and> length cs = 2
       then (CPF_Parser.rule xml2name (cs ! 0) \<guillemotright>= (\<lambda>r. CPF_Parser.rseq xml2name \<pi> r (cs ! 1)))
       else Xmlt.fail tag (XML name atts cs text)))"
hide_const (open) projected_rseq

definition rules :: "'a xmlt \<Rightarrow> ('a, string) rules xmlt" where
  "rules xml2name = Xmlt.many ''rules'' (CPF_Parser.rule xml2name) id"
hide_const (open) rules


partial_function (sum_bot)
  arith_fun :: "arithFun xmlt"
where
  [code]: "arith_fun xml = Xmlt.singleton ''arithFunction'' (Xmlt.options [ 
     (''natural'', Xmlt.change (Xmlt.nat ''natural'') Const),
     (''variable'', Xmlt.change (Xmlt.nat ''variable'') (\<lambda>n::nat. Arg (n - 1))),
     (''sum'', Xmlt.many ''sum'' arith_fun Sum),
     (''product'', Xmlt.many ''product'' arith_fun Prod),
     (''min'', Xmlt.many ''min'' arith_fun Min),
     (''max'', Xmlt.many ''max'' arith_fun Semantic_Labeling_Carrier.Max),
     (''ifEqual'', Xmlt.tuple4 ''ifEqual'' arith_fun arith_fun arith_fun arith_fun IfEqual)
   ]) id  xml"

definition
  sl_variant :: "'f xmlt \<Rightarrow> ('f,string) sl_variant xmlt"
where
  "sl_variant xml2name \<equiv> Xmlt.singleton ''model'' (Xmlt.options [
     (''finiteModel'', xml1or2many_elements ''finiteModel''
         (Xmlt.change (Xmlt.nat ''carrierSize'') (\<lambda> n :: nat. n - 1))
         (Xmlt.singleton ''tupleOrder'' (Xmlt.leaf ''pointWise'' ()) id)
         (Xmlt.triple ''interpret''
           xml2name
           (Xmlt.nat ''arity'')
           CPF_Parser.arith_fun
           (\<lambda>f a inter. ((f, a), inter)))
         (\<lambda>n model xs. if model = None then Finitelab (SL_Inter n xs) else QuasiFinitelab (SL_Inter n xs) [])),
     (''rootLabeling'',Xmlt.guard (\<lambda> x. Xml.num_children x = 0) 
            (Xmlt.leaf ''rootLabeling'' (Rootlab None))
            (Xmlt.singleton ''rootLabeling'' xml2name (\<lambda> f. Rootlab (Some (f,1))))
        )
   ]) id"
hide_const (open) sl_variant

context
  fixes C :: "'a ring" (structure)
begin
fun sum_lpoly :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a list" where
  "sum_lpoly [] ys = ys"
| "sum_lpoly xs [] = xs"
| "sum_lpoly (x # xs) (y # ys) = (x \<oplus> y) # sum_lpoly xs ys"

fun
  lpoly_of :: 
    "(nat, 'a) tpoly \<Rightarrow> string +\<^sub>\<bottom> ('a \<times> 'a list)"
where
  "lpoly_of (PNum i) = return (i, [])"
| "lpoly_of (PVar x) = return (\<zero>, replicate x \<zero> @ [\<one>])"
| "lpoly_of (PSum []) = return (\<zero>, [])"
| "lpoly_of (PSum (p # ps)) = (
     lpoly_of  p 
  \<guillemotright>= (\<lambda>(cp, ncp). lpoly_of (PSum ps)
  \<guillemotright>= (\<lambda>(cq, ncq). return (cp \<oplus> cq, sum_lpoly ncp ncq))))"
| "lpoly_of (PMult []) = return (\<one>, [])"
| "lpoly_of (PMult (p # ps)) = (
     lpoly_of p
  \<guillemotright>= (\<lambda>(cp, ncp). lpoly_of (PMult ps)
  \<guillemotright>= (\<lambda>(cq, ncq). if (Ball (set ncp) (op = \<zero>))
       then return (cp \<otimes> cq, map (op \<otimes> cp) ncq)
       else (if (Ball (set ncq) (op = \<zero>))
         then return (cp \<otimes> cq, map (\<lambda> x. x \<otimes> cq) ncp)
         else error ''cannot transform non-linear polynomial to linear polynomial''))))"

fun fit_length :: "nat \<Rightarrow> 'a list \<Rightarrow> 'a list" where
  "fit_length (Suc n) as = (case as of 
    [] \<Rightarrow> replicate (Suc n) \<zero>
  | b # bs \<Rightarrow> b # fit_length n bs)"
| "fit_length 0 _ = []"
end

partial_function (sum_bot) polynomial :: "'a xmlt \<Rightarrow> (nat, 'a) tpoly xmlt" where
  [code]: "polynomial xml2coeff x = Xmlt.singleton ''polynomial'' (Xmlt.options [
    (''coefficient'', Xmlt.singleton ''coefficient'' xml2coeff PNum), 
    (''variable'', Xmlt.change (Xmlt.nat ''variable'') (\<lambda>n. PVar (n - 1))),
    (''sum'', Xmlt.many ''sum'' (polynomial xml2coeff) PSum),
    (''product'', Xmlt.many ''product'' (polynomial xml2coeff) PMult)
  ]) id x"

definition int_coeff :: "int xmlt" where
  "int_coeff \<equiv> Xmlt.int ''integer''"
hide_const (open) int_coeff

definition vec_coeff :: "'a xmlt \<Rightarrow> 'a vec xmlt" where
  "vec_coeff xml2coeff \<equiv> Xmlt.many ''vector''
     (Xmlt.singleton ''coefficient'' xml2coeff id) id"
hide_const (open) vec_coeff

(* allow vectors and convert them to matrices by filling in zeros *)
definition mat_coeff :: "'a \<Rightarrow> 'a xmlt \<Rightarrow> 'a mat xmlt" where
  "mat_coeff ze xml2coeff \<equiv> Xmlt.options [
     (''matrix'', Xmlt.many ''matrix'' (CPF_Parser.vec_coeff xml2coeff) id),
     (''vector'', Xmlt.change
       (CPF_Parser.vec_coeff xml2coeff)
       (\<lambda>v. v # replicate (length v - 1) (replicate (length v) ze))
     )]"
hide_const (open) mat_coeff

definition arctic_coeff :: "arctic xmlt" where
  "arctic_coeff \<equiv> Xmlt.options [
    (''integer'', Xmlt.change (Xmlt.int ''integer'') Num_arc),
    (''minusInfinity'', Xmlt.leaf ''minusInfinity'' MinInfty)
  ]"
hide_const (open) arctic_coeff

definition arctic_rat_coeff :: "rat arctic_delta xmlt" where
  "arctic_rat_coeff \<equiv> Xmlt.options [
    (''integer'', Xmlt.change Xmlt.rat Num_arc_delta),
    (''rational'', Xmlt.change Xmlt.rat Num_arc_delta),
    (''minusInfinity'', Xmlt.leaf ''minusInfinity'' MinInfty_delta)
  ]"
hide_const (open) arctic_rat_coeff

datatype Domain =
  Natural nat | Integer | Arctic | Arctic_rat | Int_mat nat nat | Arc_mat nat
| Arc_rat_mat nat | Rational rat nat | Rat_mat nat nat | Mini_Alg real nat 
| Mini_Alg_mat nat nat

definition rat_domain :: "rat xmlt" where
  "rat_domain \<equiv> Xmlt.singleton ''rationals'' (Xmlt.singleton ''delta'' Xmlt.rat id) id"
hide_const (open) rat_domain

definition real :: "real xmlt" where
  "real x = (try
    Xmlt.change Xmlt.rat (real_of_rat) x
  catch (\<lambda>_. 
    Xmlt.triple ''algebraic'' (Xmlt.rat) (Xmlt.rat) (Xmlt.rat)
      (\<lambda>a b c. (real_of_rat a + real_of_rat b * sqrt (real_of_rat c))) x))"

definition real_domain :: "real xmlt" where
  "real_domain = Xmlt.singleton ''algebraicNumbers'' (Xmlt.singleton ''delta'' real id) id"
hide_const (open) real_domain

definition basic_domain :: "(nat \<Rightarrow> Domain) xmlt" where
  "basic_domain \<equiv> Xmlt.options [
    (''naturals'', Xmlt.leaf ''naturals'' Natural),
    (''integers'', Xmlt.leaf ''integers'' (\<lambda> d. Integer)),
    (''arctic'', Xmlt.singleton ''arctic'' 
      (Xmlt.singleton ''domain'' (Xmlt.options [
        (''naturals'', Xmlt.leaf ''naturals'' (\<lambda> d. Arctic)),
        (''integers'', Xmlt.leaf ''integers'' (\<lambda> d. Arctic)),
        (''rationals'', Xmlt.change CPF_Parser.rat_domain (\<lambda> r d. Arctic_rat))
      ]) id) id),
    (''rationals'', Xmlt.change CPF_Parser.rat_domain Rational),
    (''algebraicNumbers'', Xmlt.change CPF_Parser.real_domain Mini_Alg)
  ]"
hide_const (open) basic_domain

definition interpretation_type :: "Domain xmlt" where
  "interpretation_type \<equiv> (Xmlt.singleton ''type'' 
     (Xmlt.options [
       (''polynomial'', (Xmlt.pair ''polynomial'' 
       (Xmlt.singleton ''domain'' 
       (Xmlt.options [
         (''naturals'', CPF_Parser.basic_domain),
         (''integers'', CPF_Parser.basic_domain),
         (''arctic'', CPF_Parser.basic_domain),
         (''rationals'', CPF_Parser.basic_domain),
         (''algebraicNumbers'', CPF_Parser.basic_domain),
         (''matrices'',Xmlt.triple ''matrices''
           (Xmlt.nat ''dimension'') 
           (Xmlt.nat ''strictDimension'') 
           (Xmlt.singleton ''domain'' CPF_Parser.basic_domain id)
           (\<lambda>di sd domain d. case domain d of
             Natural _ \<Rightarrow> Int_mat di sd
           | Rational _ _ \<Rightarrow> Rat_mat di sd
           | Mini_Alg _ _ \<Rightarrow> Mini_Alg_mat di sd
           | Integer \<Rightarrow> Int_mat di sd
           | Arctic \<Rightarrow> Arc_mat di
           | Arctic_rat \<Rightarrow> Arc_rat_mat di
         ))
       ]) 
     id)
     (Xmlt.nat ''degree'') 
     (\<lambda>type d. type d))),
     (''matrixInterpretation'',Xmlt.triple ''matrixInterpretation''
       (Xmlt.singleton ''domain'' CPF_Parser.basic_domain id)
       (Xmlt.nat ''dimension'') 
       (Xmlt.nat ''strictDimension'') 
       (\<lambda>domain di sd. case domain 0 of
         Natural _ \<Rightarrow> Int_mat di sd
       | Integer \<Rightarrow> Int_mat di sd
       | Rational _ _  \<Rightarrow> Rat_mat di sd
       | Mini_Alg _ _  \<Rightarrow> Mini_Alg_mat di sd
       | Arctic \<Rightarrow> Arc_mat di
       | Arctic_rat \<Rightarrow> Arc_rat_mat di
     ))
   ])
 id)"
hide_const (open) interpretation_type

type_synonym ('f,'a) inter = "('f \<times> nat) \<times> 'a \<times> 'a list"
type_synonym ('f,'a) mat_inter = "('f \<times> nat) \<times> 'a mat \<times> 'a mat list"
type_synonym ('f,'a) nl_inter = "('f \<times> nat) \<times> ((nat \<times> nat) list \<times> 'a)list"

datatype 'f "interpretation" = 
  (* linear polynomials (with support of negative constants *)
  Int_linear_poly "('f,int)inter" 
| Rat_linear_poly  "('f,rat)inter"
| Arctic_linear_poly  "('f,arctic)inter"
| Arctic_rat_linear_poly  "('f,rat arctic_delta)inter"
| Real_linear_poly  "('f,real)inter"
  (* linear polynomials over matrices (with support of negative constants *)
| Int_matrix  "('f,int)mat_inter"
| Rat_matrix  "('f,rat)mat_inter"
| Arctic_matrix  "('f,arctic)mat_inter"
| Arctic_rat_matrix  "('f,rat arctic_delta)mat_inter"
| Real_matrix  "('f,real)mat_inter"
  (* non linear polynomials (without support of negative constants) *)
| Int_non_linear_poly  "('f,int)nl_inter"
| Rat_non_linear_poly  "('f,rat)nl_inter"
| Real_non_linear_poly  "('f,real)nl_inter"

fun int_linear_poly where "int_linear_poly (Int_linear_poly x) = x"
fun rat_linear_poly where "rat_linear_poly (Rat_linear_poly x) = x"
fun arctic_linear_poly where "arctic_linear_poly (Arctic_linear_poly x) = x"
fun arctic_rat_linear_poly where "arctic_rat_linear_poly (Arctic_rat_linear_poly x) = x"
fun real_linear_poly where "real_linear_poly (Real_linear_poly x) = x"
fun int_matrix where "int_matrix (Int_matrix x) = x"
fun rat_matrix where "rat_matrix (Rat_matrix x) = x"
fun arctic_matrix where "arctic_matrix (Arctic_matrix x) = x"
fun arctic_rat_matrix where "arctic_rat_matrix (Arctic_rat_matrix x) = x"
fun real_matrix where "real_matrix (Real_matrix x) = x"
fun int_non_linear_poly where "int_non_linear_poly (Int_non_linear_poly x) = x"
fun rat_non_linear_poly where "rat_non_linear_poly (Rat_non_linear_poly x) = x"
fun real_non_linear_poly where "real_non_linear_poly (Real_non_linear_poly x) = x"

(* the flag bi is to indicate that we are interested in bounded increase,
   where rationals always have to be chosen as non-linear variant *) 
abbreviation class_ring where 
  "class_ring t \<equiv> class_semiring t ()"

abbreviation mat_ring :: "('a :: semiring_1) itself \<Rightarrow> nat \<Rightarrow> 'a mat ring" where
  "mat_ring t n \<equiv> Matrix_Arith.mat_ring n ()"

definition "interpretation" :: "bool \<Rightarrow> 'a xmlt \<Rightarrow> 'a redtriple_impl xmlt" where
  "interpretation bi xml2name \<equiv> 
      (Xmlt.many1_gen ''interpretation'' 
       CPF_Parser.interpretation_type
       (\<lambda>type. case type of
          Natural deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then (\<lambda>c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.int_coeff x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly.  (Int_linear_poly ((f,a),(fst poly, fit_length c a (snd poly)))))))
                   (class_ring (TYPE(int))) else (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.int_coeff x \<guillemotright>= (\<lambda> p. return (poly_of p)))
                        (\<lambda>f a poly.  (Int_non_linear_poly ((f,a),poly))))
        | Integer \<Rightarrow> Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.int_coeff x \<guillemotright>= (\<lambda> p. return (poly_of p)))
                        (\<lambda>f a poly.  (Int_non_linear_poly ((f,a),poly)))
        | Rational d deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then (\<lambda> c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial Xmlt.rat x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly.  (Rat_linear_poly ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (class_ring (TYPE(rat))) else (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial Xmlt.rat x \<guillemotright>= (\<lambda> p. return (poly_of p)))
                        (\<lambda>f a poly.  (Rat_non_linear_poly ((f,a),poly))))
        | Mini_Alg d deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then (\<lambda> c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial real x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly.  (Real_linear_poly ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (class_ring (TYPE(real))) else (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial real x \<guillemotright>= (\<lambda> p. return (poly_of p)))
                        (\<lambda>f a poly.  (Real_non_linear_poly ((f,a),poly))))
        | Arctic \<Rightarrow> (\<lambda> c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.arctic_coeff x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly.  (Arctic_linear_poly ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (class_ring (TYPE(arctic)))
        | Arctic_rat \<Rightarrow> (\<lambda> c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.arctic_rat_coeff x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly.  (Arctic_rat_linear_poly ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (class_ring (TYPE(rat arctic_delta)))
        | Int_mat n sd \<Rightarrow> (\<lambda> c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff 0 CPF_Parser.int_coeff) x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly.  (Int_matrix ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (mat_ring (TYPE(int)) n)
        | Rat_mat n sd \<Rightarrow> (\<lambda>c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff 0 Xmlt.rat) x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly.  (Rat_matrix ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (mat_ring (TYPE(rat)) n)
        | Mini_Alg_mat n sd \<Rightarrow> (\<lambda>c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff 0 real) x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly.  (Real_matrix ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (mat_ring (TYPE(real)) n)
        | Arc_mat n \<Rightarrow> (\<lambda>c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff MinInfty CPF_Parser.arctic_coeff) x \<guillemotright>= lpoly_of c)
                        (\<lambda> f a poly. (Arctic_matrix ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (mat_ring (TYPE(arctic)) n)
        | Arc_rat_mat n \<Rightarrow> (\<lambda>c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff MinInfty_delta CPF_Parser.arctic_rat_coeff) x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly. (Arctic_rat_matrix ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (mat_ring (TYPE(rat arctic_delta)) n)
        )
      (\<lambda> type pi. case type of 
            Natural deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then Int_carrier (map int_linear_poly pi) else Int_nl_carrier (map int_non_linear_poly pi)
          | Integer \<Rightarrow> Int_nl_carrier (map int_non_linear_poly pi)
          | Rational d deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then Rat_carrier (map rat_linear_poly pi) else Rat_nl_carrier d (map rat_non_linear_poly pi)
          | Mini_Alg d deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then Real_carrier (map real_linear_poly pi) else Real_nl_carrier d (map real_non_linear_poly pi)
          | Arctic \<Rightarrow> Arctic_carrier (map arctic_linear_poly pi)
          | Arctic_rat \<Rightarrow> Arctic_rat_carrier (map arctic_rat_linear_poly pi)
          | Int_mat n sd \<Rightarrow> Int_mat_carrier n sd (map int_matrix pi)
          | Rat_mat n sd \<Rightarrow> Rat_mat_carrier n sd (map rat_matrix pi)
          | Mini_Alg_mat n sd \<Rightarrow> Real_mat_carrier n sd (map real_matrix pi)
          | Arc_mat n \<Rightarrow> Arc_mat_carrier n (map arctic_matrix pi)
          | Arc_rat_mat n \<Rightarrow> Arc_rat_mat_carrier n (map arctic_rat_matrix pi)
      ))
    "

hide_const (open) "interpretation"

definition status_precedence :: "'a xmlt \<Rightarrow> ('a status_prec_repr) xmlt" where
  "status_precedence xml2name \<equiv> Xmlt.many ''statusPrecedence'' 
  (Xmlt.tuple4 ''statusPrecedenceEntry''
      xml2name
      (Xmlt.nat ''arity'')
      (Xmlt.nat ''precedence'')
      (Xmlt.options [(''lex'', Xmlt.leaf ''lex'' Lex), (''mul'', Xmlt.leaf ''mul'' Mul)])
      (\<lambda>f a p s. (((f,a), (p,s))))
  ) id" 
hide_const (open) status_precedence

definition path_order :: "'a xmlt \<Rightarrow> 'a redtriple_impl xmlt" where
  "path_order xml2name \<equiv> xml1to2elements ''pathOrder''
   (CPF_Parser.status_precedence xml2name)
   (CPF_Parser.afs xml2name)
   (\<lambda>prec\<tau> afo. let afs = (case afo of None \<Rightarrow> [] | Some af \<Rightarrow> af)
 in
              RPO prec\<tau> afs)"
hide_const (open) path_order

definition precedence_weight :: "'a xmlt \<Rightarrow> (nat \<Rightarrow> 'a prec_weight_repr) xmlt" where
  "precedence_weight xml2name \<equiv> Xmlt.many ''precedenceWeight'' 
  (Xmlt.guard (\<lambda> x. Xml.num_children x = 4)
    (Xmlt.tuple4 ''precedenceWeightEntry''
        xml2name
        (Xmlt.nat ''arity'')
        (Xmlt.nat ''precedence'')
        (Xmlt.nat ''weight'')
        (\<lambda>f a p w. (((f,a), (p,w,None)))))
    (Xmlt.tuple5 ''precedenceWeightEntry''
        xml2name
        (Xmlt.nat ''arity'')
        (Xmlt.nat ''precedence'')
        (Xmlt.nat ''weight'')
        (Xmlt.many ''subtermCoefficientEntries''
          (Xmlt.nat ''entry'')
          id)
        (\<lambda>f a p w e. (((f,a), (p,w,Some e)))))
  ) Pair" 
hide_const (open) precedence_weight

definition knuth_bendix_order :: "'a xmlt \<Rightarrow> 'a redtriple_impl xmlt" where
  "knuth_bendix_order xml2name \<equiv> xml2to3elements ''knuthBendixOrder''
   (Xmlt.nat ''w0'')
   (CPF_Parser.precedence_weight xml2name)
   (CPF_Parser.afs xml2name)
   (\<lambda> w0 prw afo. let afs = (case afo of None \<Rightarrow> [] | Some af \<Rightarrow> af)
 in
              KBO (prw w0) afs)"
hide_const (open) knuth_bendix_order

definition redtriple :: "bool \<Rightarrow> 'a xmlt \<Rightarrow> 'a redtriple_impl xmlt" where
  "redtriple bi xml2name = 
     Xmlt.singleton ''redPair'' 
     (Xmlt.options [
  (''interpretation'', CPF_Parser.interpretation bi xml2name), 
  (''pathOrder'', CPF_Parser.path_order xml2name),
  (''knuthBendixOrder'', CPF_Parser.knuth_bendix_order xml2name)
  ]) id"
hide_const (open) redtriple

definition level_mapping :: "'a xmlt \<Rightarrow> 'a scnp_af xmlt" where
  "level_mapping xml2name = Xmlt.many ''levelMapping'' 
     (Xmlt.many2 ''levelMappingEntry''
      xml2name
      (Xmlt.nat ''arity'')
      (Xmlt.pair ''positionLevelEntry''
         (Xmlt.nat ''position'')
         (Xmlt.nat ''level'')
         Pair
      )
      (\<lambda>f a ps. ((f,a), map (\<lambda> (p,l). (if p = 0 then a else p - 1,l)) ps))
  ) id" 

definition root_redtriple :: "'a xmlt \<Rightarrow> 'a root_redtriple_impl xmlt" where
  "root_redtriple xml2name = Xmlt.singleton ''redPair'' (       
    Xmlt.triple ''scnp'' 
      (Xmlt.singleton ''status'' 
         (Xmlt.options [
             (''ms'',Xmlt.leaf ''ms'' MS_Ext),
             (''min'',Xmlt.leaf ''min'' Min_Ext),
             (''dms'',Xmlt.leaf ''dms'' Dms_Ext),
             (''max'',Xmlt.leaf ''max'' Max_Ext)
                      ])
         id)     
      (level_mapping xml2name)
      (CPF_Parser.redtriple False xml2name) 
      SCNP) id"
hide_const (open) root_redtriple
      

definition
  ordering_constraint_proof :: "bool \<Rightarrow> 'a xmlt \<Rightarrow> 'a redtriple_impl xmlt"
where
  "ordering_constraint_proof bi xml2name = 
     Xmlt.singleton ''orderingConstraintProof'' (CPF_Parser.redtriple bi xml2name) id"

definition
  root_ordering_constraint_proof :: "'a xmlt \<Rightarrow> 'a root_redtriple_impl xmlt"
where
  "root_ordering_constraint_proof xml2name = 
     Xmlt.singleton ''orderingConstraintProof'' (CPF_Parser.root_redtriple xml2name) id"

definition
  ordering_constraint_proof' :: "bool \<Rightarrow> 'a xmlt \<Rightarrow> ('a root_redtriple_impl + 'a redtriple_impl) xmlt"
where
  "ordering_constraint_proof' bi xml2name = Xmlt.choice ''reduction pair''
     [Xmlt.change (ordering_constraint_proof bi xml2name) Inr,
      Xmlt.change (root_ordering_constraint_proof xml2name) Inl]"

hide_const (open) ordering_constraint_proof
hide_const (open) root_ordering_constraint_proof
hide_const (open) ordering_constraint_proof'

definition
  id_labeling :: "(('f,'a list)lab,'a,'a list + ('f,'a list)lab list)label" 
where
  "id_labeling _ cs = Inl cs" 

definition scg_position :: "nat xmlt" where
  "scg_position \<equiv> Xmlt.nat ''position''"
hide_const (open) scg_position

definition
  scg :: "'f xmlt \<Rightarrow> (('f, string) rule \<times> ((nat \<times> nat) list \<times> (nat \<times> nat) list)) xmlt"
where
  "scg xml2name \<equiv> Xmlt.many1 ''sizeChangeGraph'' 
    (CPF_Parser.rule xml2name)
    (Xmlt.triple ''edge''
       CPF_Parser.scg_position
       (Xmlt.bool ''strict'')
       CPF_Parser.scg_position
       (\<lambda>p s q. (s,(p,q)))
    )
    (\<lambda>lr edges. (lr,(map snd (filter fst edges), map snd (filter (\<lambda> x. \<not> fst x) edges))))"
hide_const (open) scg

definition uncurry_info :: "'f xmlt \<Rightarrow> ('f, string) uncurry_info xmlt" where
  "uncurry_info xml2name \<equiv> Xmlt.tuple4 ''uncurryInformation''
    xml2name
    (Xmlt.many ''uncurriedSymbols''
      (Xmlt.many2 ''uncurriedSymbolEntry''
        xml2name
        (Xmlt.nat ''arity'')
        xml2name
        (\<lambda>f n fs. ((f, n), fs))) id)
    (Xmlt.singleton ''uncurryRules'' (CPF_Parser.rules xml2name) id) 
    (Xmlt.singleton ''etaRules'' (CPF_Parser.rules xml2name) id) 
    (\<lambda>a sml U E. (a, sml, U, E))"

definition innermostLhss :: "'a xmlt \<Rightarrow> ('a,string)term list xmlt" where
  "innermostLhss xml2name \<equiv> Xmlt.many ''innermostLhss'' 
            (CPF_Parser.term xml2name) id"

definition strategy :: "'a xmlt \<Rightarrow> ('a,string)strategy xmlt" where
  "strategy xml2name = Xmlt.singleton ''strategy'' 
     (Xmlt.options [
        (''innermost'', Xmlt.leaf ''innermost'' Innermost),
        (''innermostLhss'', Xmlt.change (innermostLhss xml2name) Innermost_Q)
                  ]) id"

definition "default_nfs_trs = False"
definition "default_nfs_dp = True"

definition "default_nfs_nt_trs = False"
definition "default_nfs_nt_dp = False"

definition
  xml2dp_input :: "'f xmlt \<Rightarrow> (bool \<times> ('f, string) rules \<times> ('f,string)strategy \<times> ('f, string) rules) xmlt"
where
  "xml2dp_input xml2name \<equiv> Xmlt.guard (\<lambda> x. Xml.num_children x = 4)
   (Xmlt.tuple4 ''dpInput'' 
         (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
         (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id)
         (strategy xml2name)  
         (Xmlt.bool ''minimal'')
         (\<lambda> r p s m. (m,p,s,r)))
   (Xmlt.triple ''dpInput'' 
         (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
         (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id)
         (Xmlt.bool ''minimal'')
         (\<lambda> r p m. (m,p,No_Strategy,r)))"

definition
  xml2dp_input' :: "bool \<Rightarrow> ('f,'l)lab xmlt \<Rightarrow> ('f,'l,string)dppLL xmlt"
where
  "xml2dp_input' termination xml2name \<equiv> Xmlt.change (xml2dp_input xml2name) 
    (\<lambda> (m,p,q,r). (if termination then default_nfs_dp else default_nfs_nt_dp, m, p, [], strategy_to_Q q r, [], r))"

partial_function (sum_bot) 
  xml2cond_constraint :: "'f xmlt \<Rightarrow> ('f,string) cond_constraint xmlt" where
  [code]: "xml2cond_constraint xml2name x = Xmlt.singleton ''conditionalConstraint''
    (Xmlt.options [
      (''all'',Xmlt.pair ''all'' (Xmlt.text ''var'') (xml2cond_constraint xml2name) CC_all),
      (''implication'',Xmlt.many1 ''implication'' (xml2cond_constraint xml2name) (xml2cond_constraint xml2name) 
        (\<lambda> c cs. let ccs = c # cs in CC_impl (take (length cs) ccs) (last ccs))),
      (''constraint'',Xmlt.triple ''constraint'' 
        (CPF_Parser.term xml2name) 
        (Xmlt.options [
          (''rewrite'',Xmlt.leaf ''rewrite'' None),
          (''strict'',Xmlt.leaf ''strict'' (Some True)),
          (''nonStrict'',Xmlt.leaf ''nonStrict'' (Some False))
          ]) 
        (CPF_Parser.term xml2name) 
        (\<lambda> s rel t. case rel of None \<Rightarrow> CC_rewr s t | Some stri \<Rightarrow> CC_cond stri (s,t)))
    ]) id x"

partial_function (sum_bot) xml2cond_constraint_prf :: "'f xmlt \<Rightarrow> ('f,string) cond_constraint_prf xmlt" where
  [code]: "xml2cond_constraint_prf xml2name x = (let cc = xml2cond_constraint xml2name in Xmlt.singleton ''conditionalConstraintProof''
    (Xmlt.options [
      (''final'',Xmlt.leaf ''final'' Final),
      (''differentConstructor'', Xmlt.singleton ''differentConstructor'' cc Different_Constructor),
      (''sameConstructor'', Xmlt.triple ''sameConstructor'' cc cc (xml2cond_constraint_prf xml2name) Same_Constructor),
      (''variableEquation'', Xmlt.tuple4 ''variableEquation'' (Xmlt.text ''var'') (CPF_Parser.term xml2name) cc (xml2cond_constraint_prf xml2name) Variable_Equation),
      (''funargIntoVar'', Xmlt.tuple5 ''funargIntoVar'' cc CPF_Parser.position (Xmlt.text ''var'') cc (xml2cond_constraint_prf xml2name) Funarg_Into_Var),
      (''simplifyCondition'', Xmlt.tuple4 ''simplifyCondition'' cc (CPF_Parser.subst xml2name) cc (xml2cond_constraint_prf xml2name) Simplify_Condition),
      (''induction'', Xmlt.triple ''induction'' cc (Xmlt.many ''conjuncts'' cc id) (Xmlt.many ''ruleConstraintProofs'' (
        Xmlt.tuple4 ''ruleConstraintProof'' 
          (CPF_Parser.rule xml2name)
          (Xmlt.many ''subtermVarEntries''
            (Xmlt.many1 ''subtermVarEntry''
              (CPF_Parser.term xml2name)
              (Xmlt.text ''var'')
              Pair)
            id)
          cc
          (xml2cond_constraint_prf xml2name)
          (\<lambda> lr rys cc p. (lr,rys,cc,p))
        ) id) Induction),
      (''deleteCondition'', Xmlt.pair ''deleteCondition'' cc(xml2cond_constraint_prf xml2name) Delete_Condition)
    ])
    id
    x)"

definition
  xml2cond_red_pair_proof :: "'f xmlt \<Rightarrow> ('f, string) cond_red_pair_prf xmlt" where
  "xml2cond_red_pair_proof xml2name \<equiv> Xmlt.tuple4 ''condRedPairProof'' 
     xml2name
     (Xmlt.nat ''before'')
     (Xmlt.nat ''after'')
     (Xmlt.many ''conditions''
       (Xmlt.triple ''condition''
         (xml2cond_constraint xml2name)
         (Xmlt.singleton ''dpSequence'' (CPF_Parser.rules xml2name) id)
         (xml2cond_constraint_prf xml2name)
         (\<lambda> c s p. (c,s,p)))
       id)
     (\<lambda> c b a ccs. Cond_Red_Pair_Prf c ccs b a)"

definition xmlIgnore :: "string \<Rightarrow> unit xmlt" where
  "xmlIgnore name xml \<equiv> if name = (Xml.tag xml)
    then return ()
    else error (''expected '' @ name @ '' element, but found '' @ show xml)"

definition bounds_type :: "boundstype xmlt" where
  "bounds_type \<equiv> Xmlt.singleton ''type'' 
     (Xmlt.options [
         (''roof'', Xmlt.leaf ''roof'' Roof),
         (''match'', Xmlt.leaf ''match'' Match)
     ])
     id"

definition bounds_bound :: "nat xmlt" where
  "bounds_bound \<equiv> Xmlt.nat ''bound''"

definition state :: "string xmlt" where
  "state = Xmlt.text ''state''"

definition final_states :: "string list xmlt" where
  "final_states = Xmlt.many ''finalStates'' state id"

definition ta_bounds_lhs :: "'a xmlt \<Rightarrow> (('a \<times> nat) \<times> string list)xmlt" where
  "ta_bounds_lhs xml2name = (Xmlt.many2 ''lhs''
      (xml2name)
      (Xmlt.nat ''height'')
      (state)
      (\<lambda>f h qs. ((f, h), qs)))"

definition ta_normal_lhs :: "'a xmlt \<Rightarrow> ('a \<times> string list)xmlt" where
  "ta_normal_lhs xml2name = (Xmlt.many1 ''lhs''
      (xml2name)
      (state)
      Pair)"

definition transition_lhs :: "'a xmlt \<Rightarrow> ('a + string) xmlt" where
  "transition_lhs xml2lhs = Xmlt.choice ''lhs'' [ 
    (Xmlt.singleton ''lhs'' state Inr),
    (Xmlt.change xml2lhs Inl)]"

definition
  transition :: "('a \<times> string list) xmlt \<Rightarrow> ((string, 'a) ta_rule + (string \<times> string)) xmlt"
where
  "transition xml2lhs = Xmlt.pair ''transition''
    (transition_lhs xml2lhs)
    (Xmlt.singleton ''rhs'' state id)
    (\<lambda>lhs q. case lhs of
      Inl (f, qs) \<Rightarrow> Inl (TA_rule f qs q)
    | Inr q' \<Rightarrow> Inr (q', q))"

definition
  transitions ::
    "('a \<times> string list) xmlt \<Rightarrow> (((string, 'a) ta_rule list) \<times> ((string \<times> string) list)) xmlt"
where
  "transitions xml2lhs = Xmlt.many ''transitions'' (transition xml2lhs)
     (\<lambda>rls. let (rules, eps) = partition (\<lambda>rl. case rl of Inl _ \<Rightarrow> True | Inr _ \<Rightarrow> False) rls;
                        ruls = map (\<lambda>rl. case rl of Inl r \<Rightarrow> r) rules;
                        ep   = map (\<lambda>rl. case rl of Inr e \<Rightarrow> e) eps
            in (ruls, ep))"

definition tree_automaton :: "('a \<times> string list) xmlt \<Rightarrow> (string, 'a) tree_automaton xmlt" where
  "tree_automaton xml2lhs \<equiv> Xmlt.pair ''treeAutomaton''
    final_states
    (transitions xml2lhs)
    (\<lambda> f (t, eps). Tree_Automaton f t eps)"

definition closed_criterion :: "string ta_relation xmlt" where
  "closed_criterion = Xmlt.singleton ''criterion'' (Xmlt.options [
      (''compatibility'', Xmlt.leaf ''compatibility'' Id_Relation),
      (''stateCompatibility'', Xmlt.singleton ''stateCompatibility'' 
         (Xmlt.many ''relation''
             (Xmlt.pair ''entry'' state state Pair)
             id) Some_Relation),
      (''decisionProcedure'', Xmlt.leaf ''decisionProcedure'' Decision_Proc)
    ]) id"

definition bounds_info :: "'a xmlt \<Rightarrow> ('a, string) bounds_info xmlt" where
  "bounds_info xml2name \<equiv> Xmlt.guard (\<lambda> x. Xml.num_children x = 4) 
    (Xmlt.tuple4 ''bounds''
      bounds_type
      bounds_bound
      final_states
      (tree_automaton (ta_bounds_lhs xml2name))
      (\<lambda> a b c d. Bounds_Info a b c d Id_Relation))
    (Xmlt.tuple5 ''bounds''
      bounds_type
      bounds_bound
      final_states
      (tree_automaton (ta_bounds_lhs xml2name))
      closed_criterion  
      Bounds_Info)
    "

type_synonym ('f,'v)i_o_strategy = "('f,'v)strategy + ('f,'v)ostrategy"

definition i_o_strategy :: "'a xmlt \<Rightarrow> ('a,string)i_o_strategy xmlt" where
  "i_o_strategy xml2name = Xmlt.singleton ''strategy'' 
     (Xmlt.options [
        (''innermost'', Xmlt.leaf ''innermost'' (Inl Innermost)),
        (''outermost'', Xmlt.leaf ''outermost'' (Inr Outermost)),
        (''innermostLhss'', Xmlt.change (innermostLhss xml2name) (\<lambda> q. Inl (Innermost_Q q)))
                  ]) id"

definition
  xml2i_o_trs_input :: "'f xmlt \<Rightarrow> (('f,string)i_o_strategy \<times> ('f, string) rules \<times> ('f, string) rules option) xmlt"
where
  "xml2i_o_trs_input xml2name \<equiv> 
   (\<lambda> x. let e = Xmlt.fail ''trsInput'' x in (case x of XML tag [] (c # cs) None \<Rightarrow> if length cs \<le> 2 \<and> tag = ''trsInput'' then (
      do { 
         r \<leftarrow> (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) c;
         (str,cs') \<leftarrow> (if cs = [] \<or> Xml.tag (hd cs) \<noteq> ''strategy'' then return (Inl No_Strategy, cs) else do { s \<leftarrow> (i_o_strategy xml2name) (hd cs); return (s,tl cs)});
         (s_opt,cs'') \<leftarrow> if (cs' = [] \<or> Xml.tag (hd cs') \<noteq> ''relativeRules'') then return (None, cs') else 
           do { s \<leftarrow> (Xmlt.singleton ''relativeRules'' (CPF_Parser.rules xml2name) id) (hd cs'); return (Some s,tl cs')};
         if cs'' = [] then return (str,r,s_opt) else e
      }
   ) else e | _ \<Rightarrow> e))" 

definition
  xml2trs_input :: "'f xmlt \<Rightarrow> (('f,string)strategy \<times> ('f, string) rules \<times> ('f, string) rules option) xmlt"
where
  "xml2trs_input xml2name x \<equiv> do {
    (ios,r,s) <- xml2i_o_trs_input xml2name x;
    case ios of Inl is \<Rightarrow> return (is,r,s)
             |  Inr _ \<Rightarrow> error ''outermost strategy not supported at this point''
  }"

definition
  xml2trs_input' :: "('f,'l)lab xmlt \<Rightarrow> ('f,'l,string)qreltrsLL xmlt" where
  "xml2trs_input' xml2name \<equiv> Xmlt.change (xml2trs_input xml2name)
    (\<lambda> (q,r,so). (default_nfs_trs, strategy_to_Q q r, r, case so of None \<Rightarrow> [] | Some s \<Rightarrow> s))"

definition
  xml2i_o_trs_input' :: "('f,'l)lab xmlt \<Rightarrow> (('f,'l,string)qreltrsLL + ('f,'l,string)otrsLL) xmlt" where
  "xml2i_o_trs_input' xml2name x \<equiv> do {
    (ios,r,so) <- xml2i_o_trs_input xml2name x;
    case ios of Inl q \<Rightarrow> return (Inl (default_nfs_trs, strategy_to_Q q r, r, case so of None \<Rightarrow> [] | Some s \<Rightarrow> s))
              | Inr q \<Rightarrow> (case so of 
                   None \<Rightarrow> return (Inr (ostrategy_to_Q q r, r)) 
                 | Some _ \<Rightarrow> error ''outermost strategy not combinable with relative rules'')
    }"

definition
  xml2trs_nt_input' :: "('f,'l)lab xmlt \<Rightarrow> ('f,'l,string)qtrsLL xmlt" where
  "xml2trs_nt_input' xml2name x \<equiv>  do {
    (ios,r,s) <- xml2i_o_trs_input xml2name x;
    case (ios,s) of (Inl q,None) \<Rightarrow> return (default_nfs_nt_trs,strategy_to_Q q r,r)
             | _ \<Rightarrow> error ''not an innermost strategy or relative rules present''}"

definition
  xml2reltrs_nt_input' :: "('f,'l)lab xmlt \<Rightarrow> ('f,'l,string)qreltrsLL xmlt" where
  "xml2reltrs_nt_input' xml2name x \<equiv>  do {
    (ios,r,s) <- xml2i_o_trs_input xml2name x;
    case (ios,s) of (Inl q,so) \<Rightarrow> return (default_nfs_nt_trs,strategy_to_Q q r,r, case so of None \<Rightarrow> [] | Some s \<Rightarrow> s)
             | _ \<Rightarrow> error ''not an innermost strategy''}"

definition 
  crule :: "'f xmlt \<Rightarrow> ('f,string)crule xmlt"
where
  "crule xml2name \<equiv> Xmlt.many1 ''conditionalRule'' (CPF_Parser.rule xml2name) (CPF_Parser.rule xml2name) Pair"

definition 
  crules :: "'f xmlt \<Rightarrow> ('f,string)crules xmlt"
where
  "crules xml2name \<equiv> Xmlt.many ''conditionalRules'' (crule xml2name) id"

definition 
  xml2ctrs_input :: "'f xmlt \<Rightarrow> ('f,string)crules xmlt"
where 
  "xml2ctrs_input xml2name \<equiv> Xmlt.singleton ''ctrsInput'' (crules xml2name) id"

hide_const (open) crule crules

definition
  xml2otrs_input :: "'f xmlt \<Rightarrow> (('f,string)ostrategy \<times> ('f, string) rules) xmlt"
where
  "xml2otrs_input xml2name x \<equiv> do {
    (ios,r,s) <- xml2i_o_trs_input xml2name x;
    case (ios,s) of (Inr os,None) \<Rightarrow> return (os,r)
             | _ \<Rightarrow> error ''not an outermost strategy or relative rules present''
  }"

definition
  xml2otrs_input' :: "'f xmlt \<Rightarrow> (('f,string)term list \<times> ('f, string) rules) xmlt"
where
  "xml2otrs_input' xml2name \<equiv> Xmlt.change (xml2otrs_input xml2name) (\<lambda> (q,r). (ostrategy_to_Q q r, r))"

definition
  xml2fptrs_input :: "'f xmlt \<Rightarrow> (('f, string) rules \<times> ('f,string) forb_pattern list) xmlt"
where
  "xml2fptrs_input xml2name \<equiv> 
     Xmlt.pair ''trsInput''
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
       (Xmlt.singleton ''strategy'' (CPF_Parser.forbidden_patterns xml2name) id)
       Pair"

definition
  xml2completion_input :: "'f xmlt \<Rightarrow> (('f, string) equation list \<times> ('f, string) rules) xmlt"
where
  "xml2completion_input xml2name \<equiv> 
     Xmlt.pair ''completionInput''
       (Xmlt.singleton ''equations'' (CPF_Parser.rules xml2name) id)
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
       Pair"

definition
  xml2equational_input :: "'f xmlt \<Rightarrow> (('f, string) equation list \<times> ('f, string) term \<times> ('f, string) term) xmlt"
where
  "xml2equational_input xml2name \<equiv>
    Xmlt.pair ''equationalReasoningInput''
      (Xmlt.singleton ''equations'' (CPF_Parser.rules xml2name) id)
      (Xmlt.pair ''equation''
        (CPF_Parser.term xml2name)
        (CPF_Parser.term xml2name)
        (Pair))
      Pair"

definition complexity_class :: "complexity_class xmlt"
  where "complexity_class \<equiv> 
       Xmlt.change (Xmlt.nat ''polynomial'') Comp_Poly"

definition signature :: "'f xmlt \<Rightarrow> ('f \<times> nat) list xmlt"
  where "signature xml2name \<equiv> Xmlt.many ''signature'' 
    (Xmlt.pair ''symbol'' xml2name (Xmlt.nat ''arity'') Pair) id"

definition complexity_measure :: "'f xmlt \<Rightarrow> ('f,string)complexity_measure xmlt"
  where "complexity_measure xml2name \<equiv> Xmlt.options [
   (''derivationalComplexity'', Xmlt.singleton ''derivationalComplexity'' (signature xml2name) Derivational_Complexity),
   (''runtimeComplexity'', Xmlt.pair ''runtimeComplexity'' (signature xml2name) (signature xml2name) Runtime_Complexity)
  ]"

definition xml2unknown_input :: "unknown_info xmlt"
  where "xml2unknown_input \<equiv> 
       Xmlt.text ''unknownInput''"

definition
  xml2complexity_input :: "'f xmlt \<Rightarrow> (('f,string)strategy \<times> ('f, string) rules \<times> ('f, string) rules option \<times> ('f,string)complexity_measure \<times> complexity_class) xmlt"
where
  "xml2complexity_input xml2name \<equiv> 
     Xmlt.triple ''complexityInput''
       (xml2trs_input xml2name)
       (complexity_measure xml2name)
       complexity_class
       (\<lambda> (q,r,s) cm cc. (q,r,s,cm,cc))"

datatype ('f,'v) input = 
  DP_input "(bool \<times> ('f, 'v) rules \<times> ('f, 'v) strategy \<times> ('f, 'v) rules)"
| TRS_input "(('f, 'v) strategy \<times> ('f, 'v) rules \<times> ('f, 'v) rules option)"
| COMP_input "(('f, 'v) equation list \<times> ('f, 'v) rules)"
| EQ_input "(('f, 'v) equation list \<times> ('f, 'v) term \<times> ('f, 'v) term)"
| CPX_input "(('f, 'v)strategy \<times> ('f, 'v) rules \<times> ('f, 'v) rules option \<times> ('f,'v)complexity_measure \<times> complexity_class)"
| OTRS_input "(('f, 'v) ostrategy \<times> ('f, 'v) rules)"
| FPTRS_input "(('f, 'v) rules  \<times> ('f, 'v) forb_pattern list)"
| CTRS_input "('f, 'v) crules"
| Unknown_input unknown_info

definition
  xml2input :: "'f xmlt \<Rightarrow> ('f,string) input xmlt"
where
  "xml2input xml2name \<equiv> Xmlt.singleton ''input'' (Xmlt.choice ''input'' [
    Xmlt.change (xml2dp_input xml2name) DP_input,
    Xmlt.change (xml2trs_input xml2name) TRS_input,
    Xmlt.change (xml2completion_input xml2name) COMP_input,
    Xmlt.change (xml2equational_input xml2name) EQ_input,
    Xmlt.change (xml2complexity_input xml2name) CPX_input,
    Xmlt.change (xml2otrs_input xml2name) OTRS_input,
    Xmlt.change (xml2fptrs_input xml2name) FPTRS_input,
    Xmlt.change (xml2ctrs_input xml2name) CTRS_input,
    Xmlt.change xml2unknown_input Unknown_input
  ]) id"

abbreviation sub_proof :: "(xml \<Rightarrow> string +\<^sub>\<bottom> ('f, 'l) lab)
    \<Rightarrow> (xml \<Rightarrow> string +\<^sub>\<bottom> 'a)
    \<Rightarrow> (xml \<Rightarrow> string +\<^sub>\<bottom> 'b)
    \<Rightarrow> (xml \<Rightarrow> string +\<^sub>\<bottom> 'c)
    \<Rightarrow> (xml \<Rightarrow> string +\<^sub>\<bottom> 'd) \<Rightarrow> xml \<Rightarrow> string +\<^sub>\<bottom> ('f, 'l, string, 'a, 'b, 'c, 'd) assm_proof"
  where "sub_proof xml2name 
  xml2trs_termination_proof 
  xml2dp_termination_proof
  xml2otrs_termination_proof 
  xml2unknown_proof x \<equiv> 
    let cs = Xml.children x
     in (if length cs \<noteq> 2 then Xmlt.fail ''subProof'' x
      else let inp = take 2 (Xml.tag (cs ! 1)) in
      (if inp = ''tr'' \<or> inp = ''re'' then 
        do {
          io_trs <- xml2i_o_trs_input' xml2name (hd cs);
          (case io_trs of
            Inl qtrs \<Rightarrow> Xmlt.change xml2trs_termination_proof (\<lambda> prf. SN_assm_proof qtrs prf :: ('f, 'l, string, 'a, 'b, 'c, 'd) assm_proof)
          | Inr otrs \<Rightarrow> Xmlt.change xml2otrs_termination_proof (SN_outermost_assm_proof otrs)) (cs ! 1)
        }
      else if inp = ''dp'' then
        Xmlt.pair ''subProof'' (xml2dp_input' True xml2name) xml2dp_termination_proof Finite_assm_proof x
      else if inp = ''un'' then
        Xmlt.pair ''subProof'' xml2unknown_input xml2unknown_proof Unknown_assm_proof x
      else Xmlt.fail ''subProof'' x))"

abbreviation dp_termination_proof where "dp_termination_proof xml2name 
  xml2trs_termination_proof 
  xml2dp_termination_proof 
  xml2otrs_termination_proof 
  xml2unknown_proof \<equiv>
  Xmlt.singleton ''dpProof'' (Xmlt.options [ 
    (''pIsEmpty'', Xmlt.leaf ''pIsEmpty'' P_is_Empty),
    (''depGraphProc'', Xmlt.many ''depGraphProc''
      (xml2nd_choice ''component'' 
        (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id)
        ''realScc''
        xml2dp_termination_proof
        (\<lambda>dps prfOpt. (prfOpt, dps)))
      Dep_Graph_Proc),
    (''redPairProc'', Xmlt.triple ''redPairProc'' 
      (CPF_Parser.ordering_constraint_proof' False xml2name) 
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof 
      Redpair_Proc),
    (''usableRulesProc'', Xmlt.pair ''usableRulesProc'' 
      (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof 
      Usable_Rules_Proc),
    (''innermostLhssRemovalProc'', Xmlt.pair ''innermostLhssRemovalProc'' 
      (innermostLhss xml2name)
      xml2dp_termination_proof 
      Q_Reduction_Proc),
    (''rewritingProc'', Xmlt.guard (\<lambda> x. Xml.num_children x = 3) 
       (Xmlt.triple ''rewritingProc'' 
         (CPF_Parser.rule xml2name)
         (CPF_Parser.rstep xml2name)
         xml2dp_termination_proof 
         (\<lambda> (s,t) (p,lr,t') prof. Rewriting_Proc None (s,t) (s,t') (s,t') lr p prof))
       (Xmlt.guard (\<lambda> x. Xml.num_children x = 4)
         (Xmlt.guard (\<lambda> x. Xml.tag (Xml.children x ! 2) = ''usableRules'')
           (Xmlt.tuple4 ''rewritingProc'' 
             (CPF_Parser.rule xml2name)
             (CPF_Parser.rstep xml2name)
             (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
             xml2dp_termination_proof 
             (\<lambda> (s,t) (p,lr,t') U prof. Rewriting_Proc (Some U) (s,t) (s,t') (s,t') lr p prof))
           (Xmlt.tuple4 ''rewritingProc'' 
             (CPF_Parser.rule xml2name)
             (CPF_Parser.rstep xml2name)
             (CPF_Parser.rule xml2name)
             xml2dp_termination_proof 
             (\<lambda> (s,t) (p,lr,t') st'' prof. Rewriting_Proc None (s,t) (s,t') st'' lr p prof)))
       (Xmlt.tuple5 ''rewritingProc'' 
         (CPF_Parser.rule xml2name)
         (CPF_Parser.rstep xml2name)
         (CPF_Parser.rule xml2name)
         (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
         xml2dp_termination_proof 
         (\<lambda> (s,t) (p,lr,t') st'' U prof. Rewriting_Proc (Some U) (s,t) (s,t') st'' lr p prof)))),
    (''narrowingProc'', Xmlt.tuple4 ''narrowingProc'' 
      (CPF_Parser.rule xml2name)
      CPF_Parser.pos
      (Xmlt.singleton ''narrowings'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof 
      Narrowing_Proc),
    (''instantiationProc'', Xmlt.triple ''instantiationProc'' 
      (CPF_Parser.rule xml2name)
      (Xmlt.singleton ''instantiations'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof 
      Instantiation_Proc),
    (''forwardInstantiationProc'', xml3to4elements ''forwardInstantiationProc'' 
      (CPF_Parser.rule xml2name)
      (Xmlt.singleton ''instantiations'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof 
      Forward_Instantiation_Proc),
    (''semlabProc'',Xmlt.guard (\<lambda> x. Xml.num_children x = 4) 
     (Xmlt.tuple4 ''semlabProc'' 
       (CPF_Parser.sl_variant xml2name) 
       (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
       xml2dp_termination_proof
       (\<lambda>sli lp. Semlab_Proc sli lp []))
     (Xmlt.tuple5 ''semlabProc'' 
       (CPF_Parser.sl_variant xml2name) 
       (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
       (innermostLhss xml2name)
       xml2dp_termination_proof
       (\<lambda>sli lp lr lq. Semlab_Proc sli lp lq lr))),
    (''subtermProc'', xml1many2elements_gen ''subtermProc''
      (CPF_Parser.proj xml2name)
      (\<lambda>\<pi>. CPF_Parser.projected_rseq xml2name (create_proj \<pi>))
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id)
      xml2dp_termination_proof
      Subterm_Criterion_Proc),
    (''redPairUrProc'', Xmlt.tuple4 ''redPairUrProc'' 
      (CPF_Parser.ordering_constraint_proof' False xml2name) 
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof
      Redpair_UR_Proc),
    (''monoRedPairUrProc'', Xmlt.tuple5 ''monoRedPairUrProc'' 
      (CPF_Parser.ordering_constraint_proof False xml2name) 
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof
      Mono_Redpair_UR_Proc),
    (''monoRedPairProc'', Xmlt.tuple4 ''monoRedPairProc'' 
      (CPF_Parser.ordering_constraint_proof False xml2name) 
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof
      Mono_Redpair_Proc),
    (''uncurryProc'', Xmlt.guard (\<lambda> x. Xml.num_children x = 4) 
      (Xmlt.tuple4 ''uncurryProc'' 
        (uncurry_info xml2name) 
        (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
        (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
        xml2dp_termination_proof
        (Uncurry_Proc None))
      (Xmlt.tuple5 ''uncurryProc'' 
        (Xmlt.nat ''applicativeTop'')
        (uncurry_info xml2name) 
        (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
        (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
        xml2dp_termination_proof
        (\<lambda>n. Uncurry_Proc (Some n)))),
    (''flatContextClosureProc'', Xmlt.tuple5 ''flatContextClosureProc''
      (Xmlt.singleton ''freshSymbol'' xml2name id)
      (CPF_Parser.flat_contexts xml2name)
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      xml2dp_termination_proof
      Fcc_Proc),
    (''switchInnermostProc'', Xmlt.pair ''switchInnermostProc''
      (wcr_proof xml2name)
      xml2dp_termination_proof
      Switch_Innermost_Proc),
    (''splitProc'', Xmlt.tuple4 ''splitProc'' 
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof
      xml2dp_termination_proof
      Split_Proc),
    (''finitenessAssumption'', Xmlt.singleton ''finitenessAssumption''
      (xml2dp_input' True xml2name)
      (\<lambda> dpp. Assume_Finite dpp [])),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
      (Xmlt.text ''description'')
      (xml2dp_input' True xml2name)
      (sub_proof xml2name
           xml2trs_termination_proof
           xml2dp_termination_proof
           xml2otrs_termination_proof 
           xml2unknown_proof)
      (\<lambda> desc. Assume_Finite)),
    (''switchToTRS'',Xmlt.singleton ''switchToTRS''
         xml2trs_termination_proof
         To_Trs_Proc),
    (''unlabProc'',Xmlt.triple ''unlabProc''
         (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
         (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
         xml2dp_termination_proof
         Unlab_Proc),
    (''generalRedPairProc'', Xmlt.guard (\<lambda> x. Xml.num_children x = 5) 
      (Xmlt.tuple5 ''generalRedPairProc'' 
      (CPF_Parser.ordering_constraint_proof True xml2name) 
      (Xmlt.singleton ''strict'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''bound'' (CPF_Parser.rules xml2name) id)
      (xml2cond_red_pair_proof xml2name)
      xml2dp_termination_proof 
      (\<lambda> rp s b c p. General_Redpair_Proc rp s b c [p]))
      (Xmlt.tuple6 ''generalRedPairProc'' 
      (CPF_Parser.ordering_constraint_proof True xml2name) 
      (Xmlt.singleton ''strict'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''bound'' (CPF_Parser.rules xml2name) id)
      (xml2cond_red_pair_proof xml2name)
      xml2dp_termination_proof 
      xml2dp_termination_proof 
      (\<lambda> rp s b c ps pb. General_Redpair_Proc rp s b c [ps,pb]))),
    (''complexConstantRemovalProc'',Xmlt.triple ''complexConstantRemovalProc''
         (CPF_Parser.term xml2name)
         (Xmlt.many ''ruleMap'' 
            (Xmlt.pair ''ruleMapEntry''
              (CPF_Parser.rule xml2name)
              (CPF_Parser.rule xml2name)
              Pair) id) 
         xml2dp_termination_proof
         (\<lambda> t rls prf. Complex_Constant_Removal_Proc (Complex_Constant_Removal_Proof t rls) prf)),
    (''sizeChangeProc'', Xmlt.many1 ''sizeChangeProc''
      (Xmlt.options [
        (''subtermCriterion'', Xmlt.leaf ''subtermCriterion'' None),
        (''reductionPair'', xml1to2elements ''reductionPair'' 
          (CPF_Parser.ordering_constraint_proof False xml2name) 
          (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
          (\<lambda>redp ur. Some (redp,ur)))
      ])
      (CPF_Parser.scg xml2name)
      (\<lambda>version. case version of
        None \<Rightarrow> Size_Change_Subterm_Proc
      | Some redp_ur \<Rightarrow> Size_Change_Redpair_Proc (fst redp_ur) (snd redp_ur))
  )]) id"

abbreviation trs_termination_proof where "trs_termination_proof xml2name 
  xml2trs_termination_proof 
  xml2dp_termination_proof 
  xml2otrs_termination_proof 
  xml2unknown_proof \<equiv>
  Xmlt.options [(''trsTerminationProof'', Xmlt.singleton ''trsTerminationProof'' (Xmlt.options [
    (''rIsEmpty'', Xmlt.leaf ''rIsEmpty'' R_is_Empty),
    (''semlab'',Xmlt.guard (\<lambda> x. Xml.num_children x = 3)
     (Xmlt.triple ''semlab'' 
       (CPF_Parser.sl_variant xml2name) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
       xml2trs_termination_proof
       (\<lambda>sli. Semlab sli []))
     (Xmlt.tuple4 ''semlab'' 
       (CPF_Parser.sl_variant xml2name) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
       (innermostLhss xml2name)
       xml2trs_termination_proof
       (\<lambda>sli lr lq. Semlab sli lq lr))),
    (''split'', Xmlt.triple ''split'' 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2trs_termination_proof
      xml2trs_termination_proof
      Split),
    (''dpTrans'', Xmlt.triple ''dpTrans'' 
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id)
      (Xmlt.bool ''markedSymbols'')
      xml2dp_termination_proof
      (\<lambda>dps _ prf. DP_Trans default_nfs_dp True dps prf)),
    (''ruleRemoval'', Xmlt.triple ''ruleRemoval''
      (CPF_Parser.ordering_constraint_proof False xml2name) 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      xml2trs_termination_proof
      Rule_Removal),
    (''bounds'', (\<lambda>xml. bounds_info xml2name xml \<guillemotright>= (return \<circ> Bounds))),
    (''stringReversal'', Xmlt.pair ''stringReversal''
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2trs_termination_proof
      (\<lambda>_. String_Reversal)),
    (''removeNonApplicableRules'', Xmlt.pair ''removeNonApplicableRules''
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2trs_termination_proof
      Remove_Nonapplicable_Rules),
    (''uncurry'', Xmlt.triple ''uncurry'' 
      (uncurry_info xml2name)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2trs_termination_proof
      Uncurry),
    (''flatContextClosure'', Xmlt.triple ''flatContextClosure''
      (CPF_Parser.flat_contexts xml2name)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      xml2trs_termination_proof
      Fcc),
    (''switchInnermost'', Xmlt.pair ''switchInnermost''
      (wcr_proof xml2name)
      xml2trs_termination_proof
      Switch_Innermost),
    (''terminationAssumption'', Xmlt.singleton ''terminationAssumption''
        (xml2trs_input' xml2name)
        (\<lambda> qtrs. Assume_SN qtrs [])),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
        (Xmlt.text ''description'')
        (xml2trs_input' xml2name)
        (sub_proof xml2name
             xml2trs_termination_proof
             xml2dp_termination_proof
             xml2otrs_termination_proof 
             xml2unknown_proof)
        (\<lambda> desc. Assume_SN))
  ]) id),
  (''relativeTerminationProof'', Xmlt.singleton ''relativeTerminationProof'' (Xmlt.options [
    (''rIsEmpty'', Xmlt.leaf ''rIsEmpty''
      R_is_Empty),
    (''sIsEmpty'', Xmlt.singleton ''sIsEmpty''
      xml2trs_termination_proof
      id),
    (''ruleRemoval'', Xmlt.tuple4 ''ruleRemoval''
      (CPF_Parser.ordering_constraint_proof False xml2name) 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      xml2trs_termination_proof
      (\<lambda>rp one two prf. Rule_Removal rp (one @ two) prf)),
    (''semlab'',Xmlt.guard (\<lambda> x. Xml.num_children x = 4) 
      (Xmlt.tuple4 ''semlab'' 
       (CPF_Parser.sl_variant xml2name) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
       xml2trs_termination_proof
       (\<lambda>sli lr lrw. Semlab sli [] (lr @ lrw)))
      (Xmlt.tuple5 ''semlab'' 
       (CPF_Parser.sl_variant xml2name) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
       (innermostLhss xml2name)
       xml2trs_termination_proof
       (\<lambda>sli lr lrw lq. Semlab sli lq (lr @ lrw)))),
    (''uncurry'', Xmlt.tuple4 ''uncurry'' 
      (uncurry_info xml2name)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2trs_termination_proof
      (\<lambda>i r s p. Uncurry i (r @ s) p)),
    (''flatContextClosure'', Xmlt.tuple4 ''flatContextClosure''
      (CPF_Parser.flat_contexts xml2name)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      xml2trs_termination_proof
      (\<lambda> fcc r s. Fcc fcc (r @ s))),
    (''stringReversal'', Xmlt.triple ''stringReversal''
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2trs_termination_proof
      (\<lambda> _ _. String_Reversal)),
    (''equalityRemoval'',Xmlt.singleton ''equalityRemoval'' 
       xml2trs_termination_proof
       Drop_Equality),
    (''relativeTerminationAssumption'',
      Xmlt.singleton ''relativeTerminationAssumption''
        (xml1to2elements ''trsInput''
          (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
          (Xmlt.singleton ''relativeRules'' (CPF_Parser.rules xml2name) id)
          (\<lambda>r rwo. case rwo of
            None \<Rightarrow> Assume_SN (default_nfs_trs, [], r, []) []
          | Some rw \<Rightarrow> Assume_SN (default_nfs_trs, [], r, rw) [])) id),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
        (Xmlt.text ''description'')
        (xml2trs_input' xml2name)
        (sub_proof xml2name
             xml2trs_termination_proof
             xml2dp_termination_proof
             xml2otrs_termination_proof 
             xml2unknown_proof)
        (\<lambda> desc. Assume_SN))
  ]) id)]"

abbreviation unknown_proof where "unknown_proof xml2name 
  xml2trs_termination_proof 
  xml2dp_termination_proof 
  xml2otrs_termination_proof 
  xml2unknown_proof \<equiv>
  Xmlt.singleton ''unknownInputProof'' (Xmlt.options [
    (''unknownAssumption'', Xmlt.singleton ''unknownAssumption''
        xml2unknown_input
        (\<lambda> u. Assume_Unknown u [])),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
        (Xmlt.text ''description'')
        xml2unknown_input
        (sub_proof xml2name
             xml2trs_termination_proof
             xml2dp_termination_proof
             xml2otrs_termination_proof 
             xml2unknown_proof)
        (\<lambda> desc. Assume_Unknown))
  ]) id"

abbreviation otrs_termination_proof where "otrs_termination_proof xml2name 
  xml2trs_termination_proof 
  xml2dp_termination_proof 
  xml2otrs_termination_proof 
  xml2unknown_proof \<equiv>
  Xmlt.singleton ''trsTerminationProof'' (Xmlt.options [
    (''terminationAssumption'', Xmlt.singleton ''terminationAssumption''
        (xml2otrs_input' xml2name)
        (\<lambda> qtrs. Assume_outermost_SN qtrs [])),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
        (Xmlt.text ''description'')
        (xml2otrs_input' xml2name)
        (sub_proof xml2name
             xml2trs_termination_proof
             xml2dp_termination_proof
             xml2otrs_termination_proof 
             xml2unknown_proof)
        (\<lambda> desc. Assume_outermost_SN))
  ]) id"

datatype termination_type = 
  DP_type 
| TRS_type 
| OTRS_type 
| Unknown_type

datatype ('f,'l)termination_result = 
  DP_result "('f, 'l, string) dp_termination_proof"
| TRS_result "('f, 'l, string) trs_termination_proof"
| OTRS_result "('f, 'l, string) otrs_termination_proof"
| Unknown_result "('f, 'l, string) unknown_proof"

fun select_DP_result where "select_DP_result (DP_result x) = x"
fun select_TRS_result where "select_TRS_result (TRS_result x) = x"
fun select_OTRS_result where "select_OTRS_result (OTRS_result x) = x"
fun select_Unknown_result where "select_Unknown_result (Unknown_result x) = x"

(*FIXME: should we set this globally for IsaFoR?*)
(* manually program mutual recursion, as at the moment of writing, it was
   not supported by partial_function *)
declare [[unify_search_bound = 400, unify_trace_bound = 400]]
partial_function (sum_bot)
  xml2_termination_proof ::
    "('f :: key, 'l :: key) lab xmlt \<Rightarrow> termination_type \<Rightarrow> ('f, 'l) termination_result xmlt"
where
  "xml2_termination_proof xml2name tt x = 
    (case tt of
      DP_type \<Rightarrow> Xmlt.change (dp_termination_proof xml2name 
        (Xmlt.change (xml2_termination_proof xml2name TRS_type) select_TRS_result)  
        (Xmlt.change (xml2_termination_proof xml2name DP_type) select_DP_result)
        (Xmlt.change (xml2_termination_proof xml2name OTRS_type) select_OTRS_result)  
        (Xmlt.change (xml2_termination_proof xml2name Unknown_type) select_Unknown_result)
        ) DP_result x
    | TRS_type \<Rightarrow> Xmlt.change (trs_termination_proof xml2name 
        (Xmlt.change (xml2_termination_proof xml2name TRS_type) select_TRS_result)
        (Xmlt.change (xml2_termination_proof xml2name DP_type) select_DP_result)
        (Xmlt.change (xml2_termination_proof xml2name OTRS_type) select_OTRS_result)  
        (Xmlt.change (xml2_termination_proof xml2name Unknown_type) select_Unknown_result)
        ) TRS_result x
    | OTRS_type \<Rightarrow> Xmlt.change (otrs_termination_proof xml2name 
        (Xmlt.change (xml2_termination_proof xml2name TRS_type) select_TRS_result)
        (Xmlt.change (xml2_termination_proof xml2name DP_type) select_DP_result)
        (Xmlt.change (xml2_termination_proof xml2name OTRS_type) select_OTRS_result)  
        (Xmlt.change (xml2_termination_proof xml2name Unknown_type) select_Unknown_result)
        ) OTRS_result x
    | Unknown_type \<Rightarrow> Xmlt.change (unknown_proof xml2name 
        (Xmlt.change (xml2_termination_proof xml2name TRS_type) select_TRS_result)
        (Xmlt.change (xml2_termination_proof xml2name DP_type) select_DP_result)
        (Xmlt.change (xml2_termination_proof xml2name OTRS_type) select_OTRS_result)  
        (Xmlt.change (xml2_termination_proof xml2name Unknown_type) select_Unknown_result)
        ) Unknown_result x
    )"

definition xml2dp_termination_proof ::
    "('f :: key, 'l :: key) lab xmlt \<Rightarrow> ('f, 'l, string) dp_termination_proof xmlt" where
  "xml2dp_termination_proof xml2name \<equiv> Xmlt.change (xml2_termination_proof xml2name DP_type) select_DP_result" 

definition xml2trs_termination_proof :: "('f :: key, 'l :: key) lab xmlt \<Rightarrow> ('f, 'l, string) trs_termination_proof xmlt" where
  "xml2trs_termination_proof xml2name = Xmlt.change (xml2_termination_proof xml2name TRS_type) select_TRS_result"

definition xml2otrs_termination_proof :: "('f :: key, 'l :: key) lab xmlt \<Rightarrow> ('f, 'l, string) otrs_termination_proof xmlt" where
  "xml2otrs_termination_proof xml2name = Xmlt.change (xml2_termination_proof xml2name OTRS_type) select_OTRS_result"

definition xml2unknown_proof :: "('f :: key, 'l :: key) lab xmlt \<Rightarrow> ('f, 'l, string) unknown_proof xmlt" where
  "xml2unknown_proof xml2name = Xmlt.change (xml2_termination_proof xml2name Unknown_type) select_Unknown_result"

text {* For generated code, eliminate construction and use direct mutual recursion without 
  @{const xml2_termination_proof} *}
lemma change_trs_termination_proof: "Xmlt.change (xml2_termination_proof xml2name TRS_type) select_TRS_result = 
  xml2trs_termination_proof xml2name" 
  by (rule ext, unfold 
    xml2trs_termination_proof_def, simp)

lemma change_otrs_termination_proof: "Xmlt.change (xml2_termination_proof xml2name OTRS_type) select_OTRS_result = 
  xml2otrs_termination_proof xml2name" 
  by (rule ext, unfold 
    xml2otrs_termination_proof_def, simp)

lemma change_dp_termination_proof: "Xmlt.change (xml2_termination_proof xml2name DP_type) select_DP_result = 
  xml2dp_termination_proof xml2name" 
  by (rule ext, unfold 
    xml2dp_termination_proof_def, simp)

lemma change_unknown_proof: "Xmlt.change (xml2_termination_proof xml2name Unknown_type) select_Unknown_result = 
  xml2unknown_proof xml2name" 
  by (rule ext, unfold 
    xml2unknown_proof_def, simp)

lemmas change_proof = 
  change_unknown_proof 
  change_dp_termination_proof
  change_trs_termination_proof
  change_otrs_termination_proof
  
lemma [code]: "xml2dp_termination_proof xml2name x = 
  dp_termination_proof xml2name
    (xml2trs_termination_proof xml2name)
    (xml2dp_termination_proof xml2name)
    (xml2otrs_termination_proof xml2name)
    (xml2unknown_proof xml2name) x"
  unfolding
    xml2dp_termination_proof_def
    Xmlt.change.simps
    xml2_termination_proof.simps
    termination_type.simps
  unfolding change_proof
  by (cases "dp_termination_proof xml2name 
    (xml2trs_termination_proof xml2name) 
    (xml2dp_termination_proof xml2name)
    (xml2otrs_termination_proof xml2name)
    (xml2unknown_proof xml2name) x", auto) 

lemma [code]: "xml2trs_termination_proof xml2name x = 
  trs_termination_proof xml2name
    (xml2trs_termination_proof xml2name)
    (xml2dp_termination_proof xml2name)
    (xml2otrs_termination_proof xml2name)
    (xml2unknown_proof xml2name) x"
  unfolding
    xml2trs_termination_proof_def
    Xmlt.change.simps
    xml2_termination_proof.simps
    termination_type.simps
  unfolding change_proof
  by (cases "trs_termination_proof xml2name 
    (xml2trs_termination_proof xml2name) 
    (xml2dp_termination_proof xml2name)
    (xml2otrs_termination_proof xml2name)
    (xml2unknown_proof xml2name) x", auto)

lemma [code]: "xml2otrs_termination_proof xml2name x = 
  otrs_termination_proof xml2name
    (xml2trs_termination_proof xml2name)
    (xml2dp_termination_proof xml2name)
    (xml2otrs_termination_proof xml2name)
    (xml2unknown_proof xml2name) x"
  unfolding
    xml2otrs_termination_proof_def
    Xmlt.change.simps
    xml2_termination_proof.simps
    termination_type.simps
  unfolding change_proof
  by (cases "otrs_termination_proof xml2name 
    (xml2trs_termination_proof xml2name) 
    (xml2dp_termination_proof xml2name)
    (xml2otrs_termination_proof xml2name)
    (xml2unknown_proof xml2name) x", auto)

lemma [code]: "xml2unknown_proof xml2name x = 
  unknown_proof xml2name
    (xml2trs_termination_proof xml2name)
    (xml2dp_termination_proof xml2name)
    (xml2otrs_termination_proof xml2name)
    (xml2unknown_proof xml2name) x"
  unfolding
    xml2unknown_proof_def
    Xmlt.change.simps
    xml2_termination_proof.simps
    termination_type.simps
  unfolding change_proof
  by (cases "unknown_proof xml2name 
    (xml2trs_termination_proof xml2name) 
    (xml2dp_termination_proof xml2name)
    (xml2otrs_termination_proof xml2name)
    (xml2unknown_proof xml2name) x", auto)
        
partial_function (sum_bot) 
  xml2complexity_proof :: "'f xmlt \<Rightarrow> ('f, string) complexity_proof xmlt"
where
  [code]: "xml2complexity_proof xml2name x = Xmlt.singleton ''complexityProof'' (Xmlt.options [
    (''ruleShifting'', Xmlt.triple ''ruleShifting''
      (CPF_Parser.ordering_constraint_proof False xml2name) 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      (xml2complexity_proof xml2name)
      Rule_Shift_Complexity),
    (''removeNonApplicableRules'', Xmlt.pair ''removeNonApplicableRules''
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      (xml2complexity_proof xml2name)
      Remove_Nonapplicable_Rules_Complexity),
    (''rIsEmpty'', Xmlt.leaf ''rIsEmpty''
      RisEmpty_Complexity)
   ]) id x"


definition nonloop :: "'f xmlt \<Rightarrow> ('f,string) non_loop_prf xmlt"
  where "nonloop xml2name \<equiv> Xmlt.tuple6 ''nonLoop''
     (CPF_Parser.pat_rule_prf xml2name)
     (CPF_Parser.subst xml2name)
     (CPF_Parser.subst xml2name)
     (Xmlt.nat ''natural'')
     (Xmlt.nat ''natural'')
     CPF_Parser.pos
     Non_Loop_Prf
"

abbreviation sub_disproof where "sub_disproof xml2name 
  xml2trs_nontermination_proof 
  xml2dp_nontermination_proof
  xml2reltrs_nontermination_proof 
  xml2unknown_disproof
  x \<equiv> 
    let cs = Xml.children x
     in (if length cs \<noteq> 2 then Xmlt.fail ''subProof'' x
      else let inp = take 2 (Xml.tag (cs ! 1)) in
      (if inp = ''tr'' then 
        Xmlt.pair ''subProof'' (xml2trs_nt_input' xml2name) xml2trs_nontermination_proof Not_SN_assm_proof x
      else if inp = ''dp'' then
        Xmlt.pair ''subProof'' (xml2dp_input' False xml2name) xml2dp_nontermination_proof Infinite_assm_proof x
      else if inp = ''re'' then
        Xmlt.pair ''subProof'' (xml2trs_input' xml2name) xml2reltrs_nontermination_proof Not_RelSN_assm_proof x
      else if inp = ''un'' then
        Xmlt.pair ''subProof'' xml2unknown_input xml2unknown_disproof Unknown_assm_proof x
      else Xmlt.fail ''subProof'' x))"

abbreviation dp_nontermination_proof where "dp_nontermination_proof 
  xml2name 
  xml2trs_nontermination_proof
  xml2dp_nontermination_proof 
  xml2reltrs_nontermination_proof
  xml2unknown_disproof
  \<equiv> Xmlt.singleton ''dpNonterminationProof'' (Xmlt.options [
    (''dpRuleRemoval'', Xmlt.choice ''dpRuleRemoval'' 
         [Xmlt.triple ''dpRuleRemoval''
            (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) Some)
            (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) Some)
            xml2dp_nontermination_proof
            DP_Rule_Removal,
          Xmlt.pair ''dpRuleRemoval''
            (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) Some)
            xml2dp_nontermination_proof
            (DP_Rule_Removal None),
          Xmlt.pair ''dpRuleRemoval''
            (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) Some)
            xml2dp_nontermination_proof
            (\<lambda> p. DP_Rule_Removal p None)
         ]),
    (''loop'', Xmlt.change (CPF_Parser.loop xml2name)
      (\<lambda>(s, rseq, \<sigma>, C). DP_Loop s rseq \<sigma> C)),
    (''nonLoop'', Xmlt.change (CPF_Parser.nonloop xml2name)
      DP_Nonloop),
    (''innermostLhssRemovalProc'', Xmlt.pair ''innermostLhssRemovalProc'' 
      (innermostLhss xml2name)
      xml2dp_nontermination_proof
      DP_Q_Reduction),
    (''innermostLhssIncreaseProc'', Xmlt.pair ''innermostLhssIncreaseProc'' 
      (innermostLhss xml2name)
      xml2dp_nontermination_proof 
      DP_Q_Increase),
    (''instantiationProc'', Xmlt.pair ''instantiationProc'' 
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id)
      xml2dp_nontermination_proof
      DP_Instantiation),
    (''narrowingProc'', Xmlt.tuple4 ''narrowingProc'' 
      (CPF_Parser.rule xml2name)
      CPF_Parser.pos
      (Xmlt.singleton ''narrowings'' (CPF_Parser.rules xml2name) id) 
      xml2dp_nontermination_proof 
      DP_Narrowing),
    (''rewritingProc'', Xmlt.guard (\<lambda> x. Xml.num_children x = 3) 
       (Xmlt.triple ''rewritingProc'' 
         (CPF_Parser.rule xml2name)
         (CPF_Parser.rstep xml2name)
         xml2dp_nontermination_proof
         (\<lambda> (s,t) (p,lr,t') prof. DP_Rewriting None (s,t) (s,t') (s,t') lr p prof))
       (Xmlt.guard (\<lambda> x. Xml.num_children x = 4)
         (Xmlt.guard (\<lambda> x. Xml.tag (Xml.children x ! 2) = ''usableRules'')
           (Xmlt.tuple4 ''rewritingProc'' 
             (CPF_Parser.rule xml2name)
             (CPF_Parser.rstep xml2name)
             (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
             xml2dp_nontermination_proof
             (\<lambda> (s,t) (p,lr,t') U prof. DP_Rewriting (Some U) (s,t) (s,t') (s,t') lr p prof))
           (Xmlt.tuple4 ''rewritingProc'' 
             (CPF_Parser.rule xml2name)
             (CPF_Parser.rstep xml2name)
             (CPF_Parser.rule xml2name)
             xml2dp_nontermination_proof 
             (\<lambda> (s,t) (p,lr,t') st'' prof. DP_Rewriting None (s,t) (s,t') st'' lr p prof)))
       (Xmlt.tuple5 ''rewritingProc'' 
         (CPF_Parser.rule xml2name)
         (CPF_Parser.rstep xml2name)
         (CPF_Parser.rule xml2name)
         (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
         xml2dp_nontermination_proof
         (\<lambda> (s,t) (p,lr,t') st'' U prof. DP_Rewriting (Some U) (s,t) (s,t') st'' lr p prof)))),
    (''switchFullStrategyProc'', Xmlt.pair ''switchFullStrategyProc'' 
      (wcr_proof xml2name)
      xml2dp_nontermination_proof 
      DP_Termination_Switch),
    (''nonterminationAssumption'', Xmlt.singleton ''nonterminationAssumption''
      (xml2dp_input' False xml2name)
      (\<lambda> qdp. DP_Assume_Infinite qdp [])),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
        (Xmlt.text ''description'')
        (xml2dp_input' False xml2name)
        (sub_disproof xml2name
             xml2trs_nontermination_proof
             xml2dp_nontermination_proof
             xml2reltrs_nontermination_proof
             xml2unknown_disproof)
        (\<lambda> desc. DP_Assume_Infinite)) 
  ]) id"

abbreviation trs_nontermination_proof where "trs_nontermination_proof 
  xml2name 
  xml2trs_nontermination_proof 
  xml2dp_nontermination_proof 
  xml2reltrs_nontermination_proof
  xml2unknown_disproof
  \<equiv> Xmlt.singleton ''trsNonterminationProof'' (Xmlt.options [
  (''variableConditionViolated'', Xmlt.leaf ''variableConditionViolated''
    TRS_Not_Well_Formed),
  (''loop'', Xmlt.change (CPF_Parser.loop xml2name)
    (\<lambda>(s, rseq, \<sigma>, C). TRS_Loop s (map (\<lambda>(x, y, _, z). (x, y, z)) rseq) \<sigma> C)),
  (''nonLoop'', Xmlt.change (CPF_Parser.nonloop xml2name)
    TRS_Nonloop),
  (''ruleRemoval'', Xmlt.pair ''ruleRemoval''
    (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
    xml2trs_nontermination_proof
    TRS_Rule_Removal),
  (''dpTrans'', Xmlt.triple ''dpTrans''
    (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id)
    (Xmlt.bool ''markedSymbols'')
    xml2dp_nontermination_proof
    (\<lambda>p _ prf. TRS_DP_Trans p prf)),
  (''stringReversal'', Xmlt.pair ''stringReversal''
    (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
    xml2trs_nontermination_proof
    (\<lambda>_. TRS_String_Reversal)),
  (''innermostLhssIncrease'', Xmlt.pair ''innermostLhssIncrease'' 
      (innermostLhss xml2name)
    xml2trs_nontermination_proof
      TRS_Q_Increase),
  (''nonterminationAssumption'', Xmlt.singleton ''nonterminationAssumption''
      (xml2trs_nt_input' xml2name)
      (\<lambda> qtrs. TRS_Assume_Not_SN qtrs [])),
  (''unknownProof'', Xmlt.many2 ''unknownProof''
      (Xmlt.text ''description'')
      (xml2trs_nt_input' xml2name)
      (sub_disproof xml2name
           xml2trs_nontermination_proof
           xml2dp_nontermination_proof
           xml2reltrs_nontermination_proof
           xml2unknown_disproof)
      (\<lambda> desc. TRS_Assume_Not_SN)) 
  ]) id"

abbreviation reltrs_nontermination_proof where "reltrs_nontermination_proof 
  xml2name 
  xml2trs_nontermination_proof 
  xml2dp_nontermination_proof 
  xml2reltrs_nontermination_proof
  xml2unknown_disproof
  \<equiv> Xmlt.singleton ''relativeNonterminationProof'' (Xmlt.options [
    (''variableConditionViolated'', Xmlt.leaf ''variableConditionViolated''
      Rel_Not_Well_Formed),
    (''loop'', Xmlt.change (CPF_Parser.loop xml2name)
      (\<lambda>(s, rseq, \<sigma>, C). Rel_Loop s rseq \<sigma> C)),
    (''trsNonterminationProof'', Xmlt.change xml2trs_nontermination_proof Rel_R_Not_SN),
    (''ruleRemoval'', Xmlt.triple ''ruleRemoval''
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) Some)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) Some)
      xml2reltrs_nontermination_proof
      Rel_Rule_Removal),
    (''nonterminationAssumption'', Xmlt.singleton ''nonterminationAssumption''
      (xml2reltrs_nt_input' xml2name)
      (\<lambda> qtrs. Rel_TRS_Assume_Not_SN qtrs [])),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
      (Xmlt.text ''description'')
      (xml2reltrs_nt_input' xml2name)
      (sub_disproof xml2name
           xml2trs_nontermination_proof
           xml2dp_nontermination_proof
           xml2reltrs_nontermination_proof
           xml2unknown_disproof)
      (\<lambda> desc. Rel_TRS_Assume_Not_SN))]) id"

abbreviation unknown_disproof where "unknown_disproof xml2name 
  xml2trs_nontermination_proof 
  xml2dp_nontermination_proof 
  xml2reltrs_nontermination_proof
  xml2unknown_disproof \<equiv>
  Xmlt.singleton ''unknownInputProof'' (Xmlt.options [
    (''unknownAssumption'', Xmlt.singleton ''unknownAssumption''
        xml2unknown_input
        (\<lambda> u. Assume_NT_Unknown u [])),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
        (Xmlt.text ''description'')
        xml2unknown_input
        (sub_disproof xml2name
             xml2trs_nontermination_proof
             xml2dp_nontermination_proof
             xml2reltrs_nontermination_proof
             xml2unknown_disproof)
        (\<lambda> desc. Assume_NT_Unknown))
  ]) id"


datatype nontermination_type = 
  Neg_DP_type 
| Neg_TRS_type 
| Neg_RelTRS_type
| Neg_Unknown_type

datatype ('f,'l)nontermination_result = 
  Neg_DP_result "('f, 'l, string) dp_nontermination_proof"
| Neg_TRS_result "('f, 'l, string) trs_nontermination_proof"
| Neg_RelTRS_result "('f, 'l, string) reltrs_nontermination_proof"
| Neg_Unknown_result "('f, 'l, string) neg_unknown_proof"

fun select_neg_DP_result where "select_neg_DP_result (Neg_DP_result x) = x"
fun select_neg_TRS_result where "select_neg_TRS_result (Neg_TRS_result x) = x"
fun select_neg_RelTRS_result where "select_neg_RelTRS_result (Neg_RelTRS_result x) = x"
fun select_neg_Unknown_result where "select_neg_Unknown_result (Neg_Unknown_result x) = x"

partial_function (sum_bot)
  xml2_nontermination_proof ::
    "('f, 'l) lab xmlt \<Rightarrow> nontermination_type \<Rightarrow> ('f, 'l)nontermination_result xmlt"
where
  "xml2_nontermination_proof xml2name tt x = 
    (case tt of
      Neg_DP_type \<Rightarrow> Xmlt.change (dp_nontermination_proof xml2name 
        (Xmlt.change (xml2_nontermination_proof xml2name Neg_TRS_type) select_neg_TRS_result)   
        (Xmlt.change (xml2_nontermination_proof xml2name Neg_DP_type) select_neg_DP_result)
        (Xmlt.change (xml2_nontermination_proof xml2name Neg_RelTRS_type) select_neg_RelTRS_result)   
        (Xmlt.change (xml2_nontermination_proof xml2name Neg_Unknown_type) select_neg_Unknown_result)
        ) Neg_DP_result x
    | Neg_TRS_type \<Rightarrow> Xmlt.change (trs_nontermination_proof xml2name 
        (Xmlt.change (xml2_nontermination_proof xml2name Neg_TRS_type) select_neg_TRS_result)
        (Xmlt.change (xml2_nontermination_proof xml2name Neg_DP_type) select_neg_DP_result)
        (Xmlt.change (xml2_nontermination_proof xml2name Neg_RelTRS_type) select_neg_RelTRS_result)   
        (Xmlt.change (xml2_nontermination_proof xml2name Neg_Unknown_type) select_neg_Unknown_result)
        ) Neg_TRS_result x
    | Neg_RelTRS_type \<Rightarrow> Xmlt.change (reltrs_nontermination_proof xml2name 
        (Xmlt.change (xml2_nontermination_proof xml2name Neg_TRS_type) select_neg_TRS_result)
        (Xmlt.change (xml2_nontermination_proof xml2name Neg_DP_type) select_neg_DP_result)
        (Xmlt.change (xml2_nontermination_proof xml2name Neg_RelTRS_type) select_neg_RelTRS_result)   
        (Xmlt.change (xml2_nontermination_proof xml2name Neg_Unknown_type) select_neg_Unknown_result)
        ) Neg_RelTRS_result x
    | Neg_Unknown_type \<Rightarrow> Xmlt.change (unknown_disproof xml2name 
        (Xmlt.change (xml2_nontermination_proof xml2name Neg_TRS_type) select_neg_TRS_result)
        (Xmlt.change (xml2_nontermination_proof xml2name Neg_DP_type) select_neg_DP_result)
        (Xmlt.change (xml2_nontermination_proof xml2name Neg_RelTRS_type) select_neg_RelTRS_result)   
        (Xmlt.change (xml2_nontermination_proof xml2name Neg_Unknown_type) select_neg_Unknown_result)
        ) Neg_Unknown_result x
    )"

definition xml2dp_nontermination_proof ::
    "('f, 'l) lab xmlt \<Rightarrow> ('f, 'l, string) dp_nontermination_proof xmlt" where
  "xml2dp_nontermination_proof xml2name \<equiv> Xmlt.change (xml2_nontermination_proof xml2name Neg_DP_type) select_neg_DP_result" 

definition xml2trs_nontermination_proof :: 
  "('f, 'l) lab xmlt \<Rightarrow> ('f, 'l, string) trs_nontermination_proof xmlt" where
  "xml2trs_nontermination_proof xml2name = Xmlt.change (xml2_nontermination_proof xml2name Neg_TRS_type) select_neg_TRS_result"

definition xml2reltrs_nontermination_proof :: 
  "('f, 'l) lab xmlt \<Rightarrow> ('f, 'l, string) reltrs_nontermination_proof xmlt" where
  "xml2reltrs_nontermination_proof xml2name = Xmlt.change (xml2_nontermination_proof xml2name Neg_RelTRS_type) select_neg_RelTRS_result"

definition xml2unknown_disproof :: "('f, 'l) lab xmlt \<Rightarrow> ('f, 'l, string) neg_unknown_proof xmlt" where
  "xml2unknown_disproof xml2name = Xmlt.change (xml2_nontermination_proof xml2name Neg_Unknown_type) select_neg_Unknown_result"

text {* For generated code, eliminate construction and use direct mutual recursion without 
  @{const xml2_nontermination_proof} *}
lemma change_trs_nontermination_proof: "Xmlt.change (xml2_nontermination_proof xml2name Neg_TRS_type) select_neg_TRS_result = 
  xml2trs_nontermination_proof xml2name" 
  by (rule ext, unfold 
    xml2trs_nontermination_proof_def, simp)

lemma change_reltrs_nontermination_proof: "Xmlt.change (xml2_nontermination_proof xml2name Neg_RelTRS_type) select_neg_RelTRS_result = 
  xml2reltrs_nontermination_proof xml2name" 
  by (rule ext, unfold 
    xml2reltrs_nontermination_proof_def, simp)

lemma change_dp_nontermination_proof: "Xmlt.change (xml2_nontermination_proof xml2name Neg_DP_type) select_neg_DP_result = 
  xml2dp_nontermination_proof xml2name" 
  by (rule ext, unfold 
    xml2dp_nontermination_proof_def, simp)

lemma change_unknown_disproof: "Xmlt.change (xml2_nontermination_proof xml2name Neg_Unknown_type) select_neg_Unknown_result = 
  xml2unknown_disproof xml2name" 
  by (rule ext, unfold 
    xml2unknown_disproof_def, simp)

lemmas change_disproof = 
  change_unknown_disproof 
  change_dp_nontermination_proof
  change_trs_nontermination_proof
  change_reltrs_nontermination_proof
  
lemma [code]: "xml2dp_nontermination_proof xml2name x = 
  dp_nontermination_proof xml2name
    (xml2trs_nontermination_proof xml2name)
    (xml2dp_nontermination_proof xml2name)
    (xml2reltrs_nontermination_proof xml2name)
    (xml2unknown_disproof xml2name) 
    x"
  unfolding
    xml2dp_nontermination_proof_def
    Xmlt.change.simps
    xml2_nontermination_proof.simps
    nontermination_type.simps
  unfolding change_disproof
  by (cases "dp_nontermination_proof xml2name
    (xml2trs_nontermination_proof xml2name) 
    (xml2dp_nontermination_proof xml2name)
    (xml2reltrs_nontermination_proof xml2name)
    (xml2unknown_disproof xml2name)
       x", auto)

lemma [code]: "xml2trs_nontermination_proof xml2name x = 
  trs_nontermination_proof xml2name
    (xml2trs_nontermination_proof xml2name)
    (xml2dp_nontermination_proof xml2name)
    (xml2reltrs_nontermination_proof xml2name)
    (xml2unknown_disproof xml2name) 
    x"
  unfolding
    xml2trs_nontermination_proof_def
    Xmlt.change.simps
    xml2_nontermination_proof.simps
    nontermination_type.simps
  unfolding change_disproof
  by (cases "trs_nontermination_proof xml2name 
    (xml2trs_nontermination_proof xml2name) 
    (xml2dp_nontermination_proof xml2name)
    (xml2reltrs_nontermination_proof xml2name)
    (xml2unknown_disproof xml2name)
       x", auto)

lemma [code]: "xml2reltrs_nontermination_proof xml2name x = 
  reltrs_nontermination_proof xml2name
    (xml2trs_nontermination_proof xml2name)
    (xml2dp_nontermination_proof xml2name)
    (xml2reltrs_nontermination_proof xml2name)
    (xml2unknown_disproof xml2name) 
    x"
  unfolding
    xml2reltrs_nontermination_proof_def
    Xmlt.change.simps
    xml2_nontermination_proof.simps
    nontermination_type.simps
  unfolding change_disproof
  by (cases "reltrs_nontermination_proof xml2name 
    (xml2trs_nontermination_proof xml2name) 
    (xml2dp_nontermination_proof xml2name)
    (xml2reltrs_nontermination_proof xml2name)
    (xml2unknown_disproof xml2name)
       x", auto)

lemma [code]: "xml2unknown_disproof xml2name x = 
  unknown_disproof xml2name
    (xml2trs_nontermination_proof xml2name)
    (xml2dp_nontermination_proof xml2name)
    (xml2reltrs_nontermination_proof xml2name)
    (xml2unknown_disproof xml2name) x"
  unfolding
    xml2unknown_disproof_def
    Xmlt.change.simps
    xml2_nontermination_proof.simps
    nontermination_type.simps
  unfolding change_disproof
  by (cases "unknown_disproof xml2name 
    (xml2trs_nontermination_proof xml2name) 
    (xml2dp_nontermination_proof xml2name)
    (xml2reltrs_nontermination_proof xml2name)
    (xml2unknown_disproof xml2name)
       x", auto)

partial_function (sum_bot)
  xml2outermost_nontermination_proof :: "'f xmlt \<Rightarrow> ('f, string) outermost_nontermination_proof xmlt"
where
  [code]: "xml2outermost_nontermination_proof xml2name x = Xmlt.singleton ''trsNonterminationProof'' (Xmlt.options [
  (''loop'', Xmlt.change (CPF_Parser.loop xml2name)
    (\<lambda>(s, rseq, \<sigma>, C). OTRS_Loop s (map (\<lambda>(x, y, _, z). (x, y, z)) rseq) \<sigma> C)),
  (''ruleRemoval'', Xmlt.pair ''ruleRemoval''
    (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
    (xml2outermost_nontermination_proof xml2name)
    OTRS_Rule_Removal)
  ]) id x"

partial_function (sum_bot)
  xml2fp_nontermination_proof :: "'f xmlt \<Rightarrow> ('f, string) fp_nontermination_proof xmlt"
where
  [code]: "xml2fp_nontermination_proof xml2name x = Xmlt.singleton ''trsNonterminationProof'' (Xmlt.options [
   (''loop'', (Xmlt.change (CPF_Parser.loop xml2name)
     (\<lambda>(s, rseq, \<sigma>, C). FPTRS_Loop s (map (\<lambda>(x, y, _, z). (x, y, z)) rseq) \<sigma> C)))
   ]
  ) id x"

definition
  xml2cr_proof :: "('f :: key,label_type)lab xmlt \<Rightarrow> ('f, label_type,string) cr_proof xmlt"
where
  "xml2cr_proof xml2name x = Xmlt.singleton ''crProof'' (Xmlt.options [
    (''wcrAndSN'', Xmlt.pair ''wcrAndSN''
         (wcr_proof xml2name)
         (xml2trs_termination_proof xml2name)
      SN_WCR),
    (''orthogonal'', Xmlt.leaf ''orthogonal'' Weakly_Orthogonal),
    (''stronglyClosed'', Xmlt.change (Xmlt.nat ''stronglyClosed'') Strongly_Closed)
  ]) id x"

definition xml2nwcr_proof :: "unit xmlt"
  where "xml2nwcr_proof \<equiv> Xmlt.singleton ''wcrDisproof'' (Xmlt.leaf ''nonJoinableCriticalPairsAuto'' ()) (\<lambda> _ . ())"

definition compatible_ta :: "string \<Rightarrow> 'f xmlt \<Rightarrow> ((string,'f)tree_automaton \<times> string ta_relation) xmlt" where
  "compatible_ta tag xml2name \<equiv> Xmlt.guard (\<lambda> x. Xml.num_children x = 1) 
    (Xmlt.singleton tag
      (tree_automaton (ta_normal_lhs xml2name))
      (\<lambda> ta. (ta, Id_Relation)))
    (Xmlt.pair tag
      (tree_automaton (ta_normal_lhs xml2name))
      closed_criterion  
      Pair)
    "

definition
  xml2ncr_proof :: "(string,label_type)lab xmlt \<Rightarrow> (string, label_type,string,string) ncr_proof xmlt"
where
  "xml2ncr_proof xml2name x \<equiv> let rew = CPF_Parser.rsteps xml2name in
    Xmlt.singleton ''crDisproof'' (Xmlt.options [
    (''nonWcrAndSN'', Xmlt.pair ''nonWcrAndSN''
         xml2nwcr_proof
         (xml2trs_termination_proof xml2name)
      (\<lambda> _. SN_NWCR)),
    (''nonJoinableFork'', Xmlt.triple ''nonJoinableFork'' rew rew 
           (Xmlt.options [
              (''distinctNormalForms'', Xmlt.leaf ''distinctNormalForms'' Diff_NFs),
              (''capNotUnif'', Xmlt.leaf ''capNotUnif'' Tcap_Not_Unif),
              (''emptyTreeAutomataIntersection'', Xmlt.pair ''emptyTreeAutomataIntersection''
                  (compatible_ta ''firstAutomaton'' xml2name)
                  (compatible_ta ''secondAutomaton'' xml2name)
                  (\<lambda> (ta1,rel1) (ta2,rel2). Tree_Aut_Intersect_Empty ta1 rel1 ta2 rel2))
           ])
           (\<lambda> (s,seq1) (_,seq2). Non_Join s seq1 seq2))          
  ]) id x"


definition
  xml2completion_proof :: "('f :: key,label_type)lab xmlt \<Rightarrow> ('f, label_type,string) completion_proof xmlt"
where
  "xml2completion_proof xml2name = Xmlt.triple ''completionProof'' 
     (wcr_proof xml2name)
     (xml2trs_termination_proof xml2name)
     (xml1to2elements ''equivalenceProof''
       (CPF_Parser.subsumption_proof xml2name)
       (CPF_Parser.subsumption_proof xml2name)
       Pair)
     (\<lambda> w t (s1,s2). SN_WCR_Eq w t s1 s2)"

definition
  xml2quasi_reductive_proof :: "('f :: key,label_type)lab xmlt \<Rightarrow> ('f, label_type,string) quasi_reductive_proof xmlt"
where
  "xml2quasi_reductive_proof xml2name \<equiv> Xmlt.singleton ''quasiReductiveProof'' (Xmlt.options [
    (''unraveling'', Xmlt.pair ''unraveling'' 
        (Xmlt.many ''unravelingInformation'' (Xmlt.many1 ''unravelingEntry''
           (CPF_Parser.crule xml2name)
           (CPF_Parser.rule xml2name)
           Pair) id)        
        (xml2trs_termination_proof xml2name)
        Unravel)
  ]) id" 

partial_function (sum_bot) xml2eq_proof :: "'a xmlt \<Rightarrow> ('a, string) eq_proof xmlt" where
  [code]: "xml2eq_proof xml2name x = Xmlt.options [
    (''refl'', Xmlt.singleton ''refl'' (CPF_Parser.term xml2name) Refl),
    (''sym'', Xmlt.singleton ''sym'' (xml2eq_proof xml2name) Sym),
    (''trans'', Xmlt.pair ''trans'' (xml2eq_proof xml2name) (xml2eq_proof xml2name) Trans),
    (''assm'', Xmlt.pair ''assm'' (CPF_Parser.rule xml2name) (CPF_Parser.subst xml2name) (\<lambda>r s. Assm r (mk_subst Var s))),
    (''cong'', Xmlt.many1 ''cong'' (xml2name) (xml2eq_proof xml2name) Cong)
  ] x"
 
definition
  xml2equational_proof :: "('f :: key, label_type) lab xmlt \<Rightarrow> ('f, label_type, string) equational_proof xmlt"
where
  "xml2equational_proof xml2name =
    Xmlt.singleton ''equationalProof'' (Xmlt.options [
      (''equationalProofTree'', Xmlt.singleton ''equationalProofTree'' (xml2eq_proof xml2name) Equational_Proof_Tree),
      (''conversion'', Xmlt.change (CPF_Parser.conversion xml2name) Conversion),
      (''completionAndNormalization'', Xmlt.pair ''completionAndNormalization''
        (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
        (xml2completion_proof xml2name)
        Completion_and_Normalization)
    ]) id"

definition
  xml2equational_disproof :: "('f :: key, label_type) lab xmlt \<Rightarrow> ('f, label_type, string) equational_disproof xmlt"
where
  "xml2equational_disproof xml2name =
    Xmlt.singleton ''equationalDisproof'' (Xmlt.options [
      (''completionAndNormalization'', Xmlt.pair ''completionAndNormalization''
        (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
        (xml2completion_proof xml2name)
        Completion_and_Normalization_Different)
    ]) id"

type_synonym lsymbol = "(string, label_type) lab"

fun
  xml2cert_problem :: "string xmlt \<Rightarrow> lsymbol xmlt \<Rightarrow> (string, label_type, string) cert_problem xmlt"
where
  "xml2cert_problem xml2uname xml2name (XML name egal cs text) =
  (if name = ''certificationProblem'' \<and> text = None \<and> length cs \<ge> 3 then do {
    inp     \<leftarrow> xml2input xml2name (cs ! 0);
    version \<leftarrow> Xmlt.text ''cpfVersion'' (cs!1);
    case inp of
      TRS_input (q,r,s_opt) \<Rightarrow>
      (case s_opt of 
        None \<Rightarrow> Xmlt.singleton ''proof'' (
        Xmlt.options [
          (''trsTerminationProof'',
            Xmlt.change (xml2trs_termination_proof xml2name) (\<lambda>p. TRS_Termination_Proof default_nfs_trs q r None p)),
          (''crProof'',
            if strategy_to_Q q r = []
              then Xmlt.change (xml2cr_proof xml2name) (\<lambda>p. TRS_Confluence_Proof False r p)
              else (\<lambda> _. error ''strategies for confluence unsupported'')),
          (''crDisproof'',
            if strategy_to_Q q r = []
              then Xmlt.change (xml2ncr_proof xml2name) (\<lambda>p. TRS_Non_Confluence_Proof False r p)
              else (\<lambda> _. error ''strategies for confluence unsupported'')),
          (''trsNonterminationProof'',
            Xmlt.change (xml2trs_nontermination_proof xml2name) (\<lambda>p. TRS_Nontermination_Proof default_nfs_trs q r p))]
      ) id (cs!2)
      | Some s \<Rightarrow> Xmlt.singleton ''proof'' (
        Xmlt.options [
          (''relativeTerminationProof'',
            Xmlt.change (xml2trs_termination_proof xml2name) (\<lambda>p. TRS_Termination_Proof default_nfs_trs q r (Some s) p)),
          (''relativeNonterminationProof'', 
            Xmlt.change (xml2reltrs_nontermination_proof xml2name) (\<lambda>p. Relative_TRS_Nontermination_Proof default_nfs_trs q r s p))]
      ) id (cs!2))
    | DP_input (m,p,q,r) \<Rightarrow> Xmlt.singleton ''proof'' (
      Xmlt.options [
        (''dpProof'',
          Xmlt.change (xml2dp_termination_proof xml2name) (\<lambda>prf. DP_Termination_Proof default_nfs_dp m p [] q [] r prf)),
        (''dpNonterminationProof'',
          Xmlt.change (xml2dp_nontermination_proof xml2name) (\<lambda>prf. DP_Nontermination_Proof default_nfs_dp m p q r prf))]
    ) id (cs!2)
    | COMP_input (e, r) \<Rightarrow> Xmlt.singleton ''proof'' (
      Xmlt.options [
        (''completionProof'',
          Xmlt.change (xml2completion_proof xml2name) (\<lambda>prf. Completion_Proof e r prf))]
    ) id (cs!2)
    | EQ_input (eqs, eq) \<Rightarrow> Xmlt.singleton ''proof'' (
      Xmlt.options [
        (''equationalProof'',
          Xmlt.change (xml2equational_proof xml2name) (\<lambda>prf. Equational_Proof eqs eq prf)),
        (''equationalDisproof'',
          Xmlt.change (xml2equational_disproof xml2name) (\<lambda>prf. Equational_Disproof eqs eq prf))
        ]
    ) id (cs!2)
    | CPX_input (q,r,s, cm, cc) \<Rightarrow> Xmlt.singleton ''proof'' (
          Xmlt.change (xml2complexity_proof xml2name) (\<lambda>prf. Complexity_Proof q r s cm cc prf)
    ) id (cs!2)
    | OTRS_input (q,r) \<Rightarrow> Xmlt.singleton ''proof'' (
          Xmlt.options [
          (''trsTerminationProof'',
            Xmlt.change (xml2otrs_termination_proof xml2name) (\<lambda>prf. Outermost_Termination_Proof q r prf)),
          (''trsNonterminationProof'',
            Xmlt.change (xml2outermost_nontermination_proof xml2name) (\<lambda>prf. Outermost_Nontermination_Proof q r prf))]
    ) id (cs!2)
    | FPTRS_input (r,p) \<Rightarrow> Xmlt.singleton ''proof'' (
          Xmlt.change (xml2fp_nontermination_proof xml2name) (\<lambda>prf. FP_Nontermination_Proof p r prf)
    ) id (cs!2)
    | CTRS_input ctrs \<Rightarrow> Xmlt.singleton ''proof'' (
          Xmlt.change (xml2quasi_reductive_proof xml2name) (\<lambda>prf. Quasi_Reductive_Proof ctrs prf)
    ) id (cs!2)
    | Unknown_input u \<Rightarrow> Xmlt.singleton ''proof'' (
          Xmlt.choice ''unknownInput'' [
            Xmlt.change (xml2unknown_proof xml2name) (\<lambda>prf. Unknown_Proof u prf),
            Xmlt.change (xml2unknown_disproof xml2name) (\<lambda>prf. Unknown_Disproof u prf)
          ]
    ) id (cs!2)
  } else error (''could not parse root element <'' @ name @ ''>''))"

partial_function (sum_bot)
  plain_name :: "lsymbol xmlt"
where
  [code]: "plain_name x = Xmlt.options [
     (''name'', Xmlt.change (Xmlt.text ''name'') UnLab),
     (''sharp'', Xmlt.singleton ''sharp'' plain_name Sharp),
     (''labeledSymbol'', Xmlt.pair ''labeledSymbol'' plain_name (
           Xmlt.options [
             (''numberLabel'', Xmlt.many ''numberLabel'' (Xmlt.nat ''number'') Inl),
             (''symbolLabel'', Xmlt.many ''symbolLabel'' plain_name Inr)
                       ])
     (\<lambda>f l. case l of
       Inl ls  \<Rightarrow> Lab f ls
     | Inr ls \<Rightarrow> FunLab f ls))
   ] x"

hide_const (open) plain_name

primrec
  xmldoc2cert_problem ::
    "xmldoc \<Rightarrow> string +\<^sub>\<bottom> (string, label_type, string) cert_problem"
where
  "xmldoc2cert_problem (XMLDOC header xml) = debug
    ''0''
    ''parsing xml to certification problem''
    (xml2cert_problem (Xmlt.text ''name'') CPF_Parser.plain_name xml)"

(*include debug information for xml-parsing*)
definition doc_of_string :: "string \<Rightarrow> string + xmldoc" where
  "doc_of_string = debug ''0'' ''parsing txt to xml'' Xml.doc_of_string"

fun
  parse_cert_problem ::
    "string \<Rightarrow> string +\<^sub>\<bottom> (string, label_type, string) cert_problem"
where
  "parse_cert_problem s =
    (case doc_of_string s of
      Inl x \<Rightarrow> error x
    | Inr y \<Rightarrow> xmldoc2cert_problem y)"

declare parse_cert_problem.simps [simp del]

(* ... to xml *)
fun xml_lab :: "('f :: show, ('l :: show)list)lab \<Rightarrow> xml" where
  "xml_lab (UnLab x) = XML ''name'' [] [] (Some (show x))"
| "xml_lab (Sharp x) = XML ''sharp'' [] [xml_lab x] None"
| "xml_lab (FunLab x l) = XML ''labeledSymbol'' [] [xml_lab x, XML ''symbolLabel'' [] (map xml_lab l) None] None"
| "xml_lab (Lab x l) = XML ''labeledSymbol'' [] [xml_lab x, XML ''numberLabel'' [] (map (\<lambda> n. XML ''number'' [] [] (Some (show n))) l) None] None"

definition xml_signature :: "(('f :: show,('l :: show)list)lab \<times> nat)list \<Rightarrow> xml" where
  "xml_signature fs \<equiv> XML ''signature'' [] (map (\<lambda> (f,n). XML ''symbol'' [] [xml_lab f, XML ''arity'' [] [] (Some (shows n Nil))] None) fs) None"

fun xml_complexity_measure :: "(('f :: show,('l :: show)list)lab,'v)complexity_measure \<Rightarrow> xml" where
  "xml_complexity_measure (Derivational_Complexity F) = XML ''derivationalComplexity'' [] [xml_signature F] None"
 | "xml_complexity_measure (Runtime_Complexity C D) = XML ''runtimeComplexity'' [] [xml_signature C, xml_signature D] None"

fun xml_complexity_class :: "complexity_class \<Rightarrow> xml" where
  "xml_complexity_class (Comp_Poly n) = XML ''polynomial'' [] [] (Some (shows n Nil))"

definition xml_pos :: "pos \<Rightarrow> xml" where
  "xml_pos p \<equiv>  XML ''positionInTerm'' [] (map (\<lambda> i. XML ''position'' [] [] (Some (show (Suc i)))) (to_list p)) None"

fun xml_term :: "(('f :: show,('l :: show)list)lab, 'v :: show)term \<Rightarrow> xml" where
  "xml_term (Var x) = XML ''var'' [] [] (Some (show x))" 
| "xml_term (Fun f ts) = XML ''funapp'' [] 
    (xml_lab f # 
      (map (\<lambda> t. XML ''arg'' [] [xml_term t] None) ts)) 
      None"

definition xml_rule :: "(('f :: show,('l :: show)list)lab,'c :: show) rule \<Rightarrow> xml" where
  "xml_rule \<equiv> \<lambda> (l, r). XML ''rule'' [] [
    XML ''lhs'' [] [xml_term l] None,         
    XML ''rhs'' [] [xml_term r] None
  ] None"

definition xml_rules :: "string \<Rightarrow> (('f :: show,('l :: show)list)lab,'c :: show) rules \<Rightarrow> xml" where
  "xml_rules tag rls = XML tag Nil [XML ''rules'' [] (map xml_rule rls) None] None"

definition xml_crule :: "(('f :: show,('l :: show)list)lab,'c :: show) crule \<Rightarrow> xml" where
  "xml_crule \<equiv> \<lambda> (c, rs). XML ''conditionalRule'' [] (map xml_rule (c # rs)) None"

definition xml_forbidden_pattern :: "(('f :: show,('l :: show) list)lab,'v :: show) forb_pattern \<Rightarrow> xml" where
  "xml_forbidden_pattern \<equiv> \<lambda> (C,t,l). XML ''forbiddenPattern'' [] [
    xml_term C\<langle>t\<rangle>,
    xml_pos (hole_pos C),
    (case l of Forbidden_Patterns.H \<Rightarrow> XML ''here'' [] [] None
             | Forbidden_Patterns.A \<Rightarrow> XML ''above'' [] [] None
             | Forbidden_Patterns.B \<Rightarrow> XML ''below'' [] [] None)
   ] None"

fun xml_Q :: "(('f :: show,('l :: show) list)lab,'v :: show)i_o_strategy \<Rightarrow> xml list"
  where "xml_Q (Inl No_Strategy) = []"
     |  "xml_Q (Inl Innermost) = [XML ''strategy'' [] [XML ''innermost'' [] [] None] None]"
     |  "xml_Q (Inr Outermost) = [XML ''strategy'' [] [XML ''outermost'' [] [] None] None]"
     |  "xml_Q (Inl (Innermost_Q Q)) = [XML ''strategy'' [] 
                                   [XML ''innermostLhss'' [] (map xml_term Q) None] None]"

fun xml_trs_input :: "(('f :: show,('l :: show) list)lab,'v :: show)i_o_strategy \<Rightarrow> (('f,'l list)lab,'v)rules
  \<Rightarrow> (('f,'l list)lab,'v)rules option \<Rightarrow> xml"
  where "xml_trs_input Q R None =  XML ''trsInput'' [] ([xml_rules ''trs'' R] @ xml_Q Q) None"
  |     "xml_trs_input Q R (Some S) = XML ''trsInput'' [] (xml_rules ''trs'' R # xml_Q Q @ [xml_rules ''relativeRules'' S]) None"

definition xml_ctrs_input :: "(('f :: show,('l :: show) list)lab,'v :: show)crules \<Rightarrow> xml"
  where "xml_ctrs_input ctrs =  XML ''ctrsInput'' [] [XML ''conditionalRules'' [] (map xml_crule ctrs) None] None"

definition xml_fp_trs_input :: "((('f :: show,('l :: show) list)lab,'v :: show) forb_pattern) list \<Rightarrow> (('f,'l list) lab,'v) rules \<Rightarrow> xml "
  where "xml_fp_trs_input P R = XML ''trsInput'' [] [
          xml_rules ''trs'' R, 
          XML ''strategy'' [] [XML ''forbiddenPatterns'' [] (map xml_forbidden_pattern P) None] None
         ] None"

fun xml_cert_problem :: "('f :: show, ('l :: show) list, 'v :: show) cert_problem \<Rightarrow> xml" where
  "xml_cert_problem (TRS_Termination_Proof nfs Q R S_o _) = xml_trs_input (Inl Q) R S_o"
| "xml_cert_problem (TRS_Nontermination_Proof nfs Q R _) = xml_trs_input (Inl Q) R None"
| "xml_cert_problem (Outermost_Nontermination_Proof Q R _) = xml_trs_input (Inr Q) R None"
| "xml_cert_problem (Outermost_Termination_Proof Q R _) = xml_trs_input (Inr Q) R None"
| "xml_cert_problem (FP_Nontermination_Proof P R _) = xml_fp_trs_input P R"
| "xml_cert_problem (Relative_TRS_Nontermination_Proof nfs Q R S _) = xml_trs_input (Inl Q) R (Some S)"
| "xml_cert_problem (DP_Termination_Proof nfs m P Pw Q R Rw _) =
    XML ''dpInput'' [] (xml_rules ''trs'' Rw # xml_rules ''dps'' P # xml_Q (Inl Q) @ [XML ''minimal'' [] [] (Some (if m then ''true'' else ''false''))]) None"
| "xml_cert_problem (DP_Nontermination_Proof nfs m P Q R _) =
    XML ''dpInput'' [] (xml_rules ''trs'' R # xml_rules ''dps'' P # xml_Q (Inl Q) @ [XML ''minimal'' [] [] (Some (if m then ''true'' else ''false''))]) None"
| "xml_cert_problem (TRS_Confluence_Proof nfs R _) = xml_trs_input (Inl No_Strategy) R None"
| "xml_cert_problem (TRS_Non_Confluence_Proof nfs R _) = xml_trs_input (Inl No_Strategy) R None"
| "xml_cert_problem (Completion_Proof E R _) =
    XML ''completionInput'' [] [xml_rules ''equations'' E, xml_rules ''trs'' R] None"
| "xml_cert_problem (Equational_Proof E eq _) =
    XML ''equationalReasoningInput'' [] [xml_rules ''equations'' E, XML ''equation'' [] [xml_term (fst eq), xml_term (snd eq)] None] None"
| "xml_cert_problem (Equational_Disproof E eq _) =
    XML ''equationalReasoningInput'' [] [xml_rules ''equations'' E, XML ''equation'' [] [xml_term (fst eq), xml_term (snd eq)] None] None"
| "xml_cert_problem (Complexity_Proof Q R S cm cc _) =
    XML ''complexityInput'' [] [xml_trs_input (Inl Q) R S, xml_complexity_measure cm, xml_complexity_class cc] None"
| "xml_cert_problem (Quasi_Reductive_Proof ctrs _) = xml_ctrs_input ctrs"
| "xml_cert_problem (Unknown_Proof u _) = XML ''unknownInput'' [] [] (Some u)"
| "xml_cert_problem (Unknown_Disproof u _) = XML ''unknownInput'' [] [] (Some u)"

end

