theory PCP
imports CP
begin

subsection {* Prime Critical Pairs *}

definition "PCP R =
  {(replace_at (fst r') p (snd r) \<cdot> \<sigma>, snd r' \<cdot> \<sigma>) | r p r' \<sigma>.
    overlap R R r p r' \<and>
    \<sigma> = the_mgu (fst r) (fst r' |_ p) \<and>
    (\<forall>u \<lhd> fst r \<cdot> \<sigma>. u \<in> NF (rstep R))}"

lemma PCP_subset_CP:
  "PCP R \<subseteq> CP R"
  by (force simp: PCP_def CP2_def)

text {*A (non-empty) peak that is joinable or a prime critical pair.*}
definition "nabla R s t u \<longleftrightarrow>
  (s, t) \<in> (rstep R)\<^sup>+ \<and> (s, u) \<in> (rstep R)\<^sup>+ \<and>
  ((t, u) \<in> (rstep R)\<^sup>\<down> \<or> (t, u) \<in> (rstep (PCP R))\<^sup>\<leftrightarrow>)"

lemma nabalaI_join [intro]:
  assumes "(t, u) \<in> (rstep R)\<^sup>\<down>"
    and "(s, t) \<in> (rstep R)\<^sup>+" and "(s, u) \<in> (rstep R)\<^sup>+"
  shows "nabla R s t u"
  using assms by (auto simp: nabla_def)

lemma nablaI_rstep_PCP [intro]:
  assumes "(t, u) \<in> (rstep (PCP R))\<^sup>\<leftrightarrow>"
    and "(s, t) \<in> (rstep R)\<^sup>+" and "(s, u) \<in> (rstep R)\<^sup>+"
  shows "nabla R s t u"
  using assms by (auto simp: nabla_def)

lemma nabla_refl_on_rstep_trancl:
  "(s, t) \<in> (rstep R)\<^sup>+ \<Longrightarrow> nabla R s t t"
  by auto

lemma nabla_imp_nabla_refl:
  assumes "nabla R s t u"
  shows "nabla R s u u"
proof -
  have "(s, u) \<in> (rstep R)\<^sup>+" using assms by (auto simp: nabla_def)
  from nabla_refl_on_rstep_trancl [OF this]
    show ?thesis .
qed

lemma nabla_ctxt:
  assumes "nabla R s t u"
  shows "nabla R (C\<langle>s\<rangle>) (C\<langle>t\<rangle>) (C\<langle>u\<rangle>)"
  using assms by (auto simp: nabla_def dest: trancl_rstep_ctxt_closed)

lemma nabla_subst:
  assumes "nabla R s t u"
  shows "nabla R (s \<cdot> \<sigma>) (t \<cdot> \<sigma>) (u \<cdot> \<sigma>)"
  using assms by (auto simp: nabla_def dest: rsteps_subst_closed)

lemma overlap_imp_rstep:
  assumes "overlap R R' r p r'"
  shows "r \<in> rstep R" and "r' \<in> rstep R'"
  apply (insert assms)
  apply (case_tac [!] r, case_tac [!] r')
  apply (auto simp: overlap_def)
  by (metis perm_rstep_conv rstep.id surj_pair)+

lemma rstep_pos_subst:
  assumes "(s, t) \<in> rstep_pos R p"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> rstep_pos R p"
using assms
proof (cases)
  case (rule l r \<sigma>')
  with rstep_pos.intros [OF this(2), of p "s \<cdot> \<sigma>" "\<sigma>' \<circ>s \<sigma>"]
    show ?thesis by (auto simp: ctxt_of_pos_term_subst)
qed

lemma rstep_pos_permute [simp]:
  fixes \<pi> :: "('v :: infinite) perm"
  shows "rstep_pos (\<pi> \<bullet> R) p = rstep_pos R p"
proof
  show "rstep_pos (\<pi> \<bullet> R) p \<subseteq> rstep_pos R p"
  proof (rule subrelI)
    fix s t
    assume "(s, t) \<in> rstep_pos (\<pi> \<bullet> R) p"
    then show "(s, t) \<in> rstep_pos R p"
    proof (cases)
      fix l r \<sigma>
      assume "(l, r) \<in> \<pi> \<bullet> R"
        and [simp]: "s |_ p = l \<cdot> \<sigma>" "t = (ctxt_of_pos_term p s)\<langle>r \<cdot> \<sigma>\<rangle>"
        and p: "p \<in> poss s"
      then have "(-\<pi> \<bullet> l, -\<pi> \<bullet> r) \<in> -\<pi> \<bullet> \<pi> \<bullet> R" by simp
      then have "(-\<pi> \<bullet> l, -\<pi> \<bullet> r) \<in> R" by simp
      from rstep_pos.intros [OF this p, of "sop \<pi> \<circ>s \<sigma>"]
        show "(s, t) \<in> rstep_pos R p" by simp
    qed
  qed
next
  show "rstep_pos R p \<subseteq> rstep_pos (\<pi> \<bullet> R) p"
  proof (rule subrelI)
    fix s t
    assume "(s, t) \<in> rstep_pos R p"
    then show "(s, t) \<in> rstep_pos (\<pi> \<bullet> R) p"
    proof (cases)
      fix l r \<sigma>
      assume "(l, r) \<in> R" and [simp]: "s |_ p = l \<cdot> \<sigma>" "t = (ctxt_of_pos_term p s)\<langle>r \<cdot> \<sigma>\<rangle>"
        and p: "p \<in> poss s"
      then have "(\<pi> \<bullet> l, \<pi> \<bullet> r) \<in> \<pi> \<bullet> R" by simp
      from rstep_pos.intros [OF this p, of "sop (-\<pi>) \<circ>s \<sigma>"]
        show "(s, t) \<in> rstep_pos (\<pi> \<bullet> R) p" by simp
    qed
  qed
qed

lemma rstep_pos_subset_permute_rstep_pos:
  shows "rstep_pos R p \<subseteq> \<pi> \<bullet> rstep_pos R p"
proof (rule subrelI)
  fix s t
  assume "(s, t) \<in> rstep_pos R p"
  then have "(s, t) \<in> rstep_pos (\<pi> \<bullet> R) p" by auto
  then show "(s, t) \<in> \<pi> \<bullet> rstep_pos R p"
  proof (cases)
    case (rule l r \<sigma>)
    then have p: "p \<in> poss (-\<pi> \<bullet> s)"
      and *: "- \<pi> \<bullet> s |_ p = -\<pi> \<bullet> l \<cdot> sop \<pi> \<circ>s \<sigma> \<circ>s sop (-\<pi>)" by (auto simp: eqvt [symmetric])
    have "(l, r) \<in> \<pi> \<bullet> R" by fact
    then have "(\<pi> \<bullet> (-\<pi>) \<bullet> l, \<pi> \<bullet> (-\<pi>) \<bullet> r) \<in> \<pi> \<bullet> R" by simp+
    then have "(-\<pi> \<bullet> l, -\<pi> \<bullet> r) \<in> R" by simp
    from rstep_pos.intros [OF this p *] and p
      show ?thesis unfolding rule by (auto simp: eqvt [symmetric])
  qed
qed

lemma rstep_pos_eqvt [eqvt]:
  fixes \<pi> :: "('v :: infinite) perm"
  shows "\<pi> \<bullet> rstep_pos R p = rstep_pos (\<pi> \<bullet> R) p"
proof
  show "rstep_pos (\<pi> \<bullet> R) p \<subseteq> \<pi> \<bullet> rstep_pos R p"
    using rstep_pos_subset_permute_rstep_pos by simp
next
  from rstep_pos_subset_permute_rstep_pos [of R p "-\<pi>"]
    have "\<pi> \<bullet> rstep_pos R p \<subseteq> \<pi> \<bullet> -\<pi> \<bullet> rstep_pos R p" by (metis trs_pt.permute_set_subset)
  then show "\<pi> \<bullet> rstep_pos R p \<subseteq> rstep_pos (\<pi> \<bullet> R) p" by simp
qed

lemma perm_rstep_pos_conv [simp]:
  "\<pi> \<bullet> r \<in> rstep_pos R p \<longleftrightarrow> r \<in> rstep_pos R p"
  by (metis rstep_pos_eqvt rstep_pos_permute trs_pt.mem_permute_iff)

lemma rstep_pos_rule:
  assumes "(l, r) \<in> R"
  shows "(l, r) \<in> rstep_pos R \<epsilon>"
  using rstep_pos.intros [OF assms, of \<epsilon> l Var] by simp

lemma overlap_imp_rstep_pos_epsilon:
  fixes R :: "('f, 'v :: infinite) trs"
  assumes "overlap R R' r p r'"
  shows "r \<in> rstep_pos R \<epsilon>" and "r' \<in> rstep_pos R' \<epsilon>"
proof -
  from assms [unfolded overlap_def] obtain \<pi>\<^sub>1 \<pi>\<^sub>2 l\<^sub>1 r\<^sub>1 l\<^sub>2 r\<^sub>2 and \<sigma> :: "('f, 'v) subst"
    where p: "p \<in> poss l\<^sub>2"
      and [simp]: "r = (l\<^sub>1, r\<^sub>1)" "r' = (l\<^sub>2, r\<^sub>2)"
      and 1: "(l\<^sub>1, r\<^sub>1) \<in> -\<pi>\<^sub>1 \<bullet> R" and 2: "(l\<^sub>2, r\<^sub>2) \<in> -\<pi>\<^sub>2 \<bullet> R'"
      by (cases r, cases r') (fastforce dest: funposs_imp_poss)
  from rstep_pos.intros [OF 1, of \<epsilon> l\<^sub>1 Var]
    show "r \<in> rstep_pos R \<epsilon>" by simp
  from rstep_pos.intros [OF 2, of \<epsilon> l\<^sub>2 Var]
    show "r' \<in> rstep_pos R' \<epsilon>" by simp
qed

lemma overlap_source_eq:
  fixes R :: "('f, 'v :: infinite) trs" and r p r'
  defines "\<sigma> \<equiv> the_mgu (fst r) (fst r' |_ p)"
  assumes "overlap R R' r p r'"
  shows "p \<in> poss (fst r')" (is ?A)
    and "fst r' \<cdot> \<sigma> = replace_at (fst r') p (fst r) \<cdot> \<sigma>" (is ?B)
proof -
  from assms obtain \<tau> :: "('f, 'v) subst"
    where p: "p \<in> poss (fst r')"
    and "fst r \<cdot> \<tau> = (fst r' |_ p) \<cdot> \<tau>"
    by (auto simp: overlap_def dest: funposs_imp_poss)
  from the_mgu [OF this(2)] and assms
    have "fst r \<cdot> \<sigma> = (fst r' |_ p) \<cdot> \<sigma>" by simp
  with p show ?B by (metis ctxt_supt_id subst_apply_term_ctxt_apply_distrib)
  show ?A by fact
qed

lemma overlap_rstep_pos_right:
  fixes R :: "('f, 'v :: infinite) trs" and r p r'
  defines "\<sigma> \<equiv> the_mgu (fst r) (fst r' |_ p)"
  assumes ol: "overlap R R' r p r'"
  shows "(fst r' \<cdot> \<sigma>, snd r' \<cdot> \<sigma>) \<in> rstep_pos R' \<epsilon>"
  using overlap_imp_rstep_pos_epsilon(2) [OF ol]
    and overlap_source_eq [OF ol, folded \<sigma>_def]
    and rstep_pos_subst [of "fst r'" "snd r'" R' \<epsilon> \<sigma>]
    by simp

lemma rstep_pos_supt:
  assumes "(s, t) \<in> rstep_pos R p"
    and q: "q \<in> poss u" and u: "u |_ q = s"
  shows "(u, (ctxt_of_pos_term q u)\<langle>t\<rangle>) \<in> rstep_pos R (q <#> p)"
using assms
proof (cases)
  case (rule l r \<sigma>)
  with q and u have "(q <#> p) \<in> poss u" and "u |_ (q <#> p) = l \<cdot> \<sigma>" by auto
  with rstep_pos.rule [OF rule(2) this] show ?thesis
    unfolding rule by (auto simp: ctxt_of_pos_term_append u)
qed

lemma overlap_rstep_pos_left:
  fixes R :: "('f, 'v :: infinite) trs" and r p r'
  defines "\<sigma> \<equiv> the_mgu (fst r) (fst r' |_ p)"
  assumes ol: "overlap R R' r p r'"
  shows "(fst r' \<cdot> \<sigma>, (ctxt_of_pos_term p (fst r'))\<langle>snd r\<rangle> \<cdot> \<sigma>) \<in> rstep_pos R p"
  using overlap_imp_rstep_pos_epsilon(1) [OF ol]
    and overlap_source_eq [OF ol, folded \<sigma>_def]
    and rstep_pos_supt [of "fst r \<cdot> \<sigma>" "snd r \<cdot> \<sigma>"  R \<epsilon> p "fst r' \<cdot> \<sigma>"]
    apply (cases r, cases r')
    apply auto
    by (metis ctxt_of_pos_term_subst poss_imp_subst_poss rstep_pos_subst subt_at_ctxt_of_pos_term)

lemma rrstep_rstep_pos_conv:
  "rrstep R = rstep_pos R \<epsilon>"
  by (auto simp: rrstep_def rstep_pos_rstep_r_p_s_conv)

lemma all_subt_conv:
  "(\<forall>s \<lhd> t. P s) \<longleftrightarrow> (\<forall>p\<in>poss t. p \<noteq> \<epsilon> \<longrightarrow> P (t |_ p))"
  by (metis pos.distinct(1) subt_at_id_imp_eps subt_at_imp_supteq' suptI supt_imp_subt_at_nepos)

lemma all_subst_subt_at_conv:
  assumes "p \<in> poss t"
  shows "(\<forall>s \<lhd> t |_ p. P s) \<longleftrightarrow> (\<forall>q\<in>poss t. q > p \<longrightarrow> P (t |_ q))"
  using assms by (auto simp: all_subt_conv) (metis less_pos_def' poss_append_poss subt_at_append)+

lemma rrstep_imp_rstep:
  assumes "(s, t) \<in> rrstep R"
  shows "(s, t) \<in> rstep R"
  using assms by (auto simp: rrstep_def rstep_iff_rstep_r_p_s)

lemma not_NF_rstep_imp_subteq_not_NF_rrstep:
  assumes "s \<notin> NF (rstep R)"
  shows "\<exists>t \<unlhd> s. t \<notin> NF (rrstep R)"
proof -
  from assms obtain u where "(s, u) \<in> rstep R" by auto
  then obtain l r C \<sigma> where "(l, r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and u: "u = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  then have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> rrstep R" and "l \<cdot> \<sigma> \<unlhd> s" by auto
  then show ?thesis by blast
qed

lemma all_subt_NF_rrstep_iff_all_subt_NF_rstep:
  "(\<forall>s \<lhd> t. s \<in> NF (rrstep R)) \<longleftrightarrow> (\<forall>s \<lhd> t. s \<in> NF (rstep R))"
  apply (auto dest: rrstep_imp_rstep)
  apply (rule ccontr)
  apply (frule not_NF_rstep_imp_subteq_not_NF_rrstep)
  by (auto dest: supt_supteq_trans)

lemma rstep_pos_imp_lowest_or_lower:
  assumes "(s, t) \<in> rstep_pos R p"
  shows "(\<forall>u \<lhd> s |_ p. u \<in> NF (rrstep R)) \<or> (\<exists>v. \<exists>q>p. (s, v) \<in> rstep_pos R q)"
proof (cases "\<forall>u \<lhd> s |_ p. u \<in> NF (rrstep R)")
  case True
  then show ?thesis by blast
next
 from assms have p: "p \<in> poss s" by (cases)
  case False
  from this [unfolded all_subst_subt_at_conv [OF p]] obtain q
    where "q \<in> poss s" and "q > p" and "s |_ q \<notin> NF (rrstep R)" by auto
  moreover then obtain v where "(s, v) \<in> rstep_pos R q"
    unfolding rrstep_rstep_pos_conv NF_def
    by (auto) (metis rrstep_imp_rule_subst rrstep_rstep_pos_conv rstep_pos.intros)
  ultimately show ?thesis by blast
qed

lemma not_in_poss_imp_NF_rstep_pos [simp]:
  assumes "p \<notin> poss s"
  shows "s \<in> NF (rstep_pos R p)"
  using assms by (auto simp: NF_def elim: rstep_pos.cases)

lemma Var_rstep_imp_rstep_pos_epsilon:
  assumes "(Var x, t) \<in> rstep R"
  shows "(Var x, t) \<in> rstep_pos R \<epsilon>"
  using assms by (metis Var_supt nrrstep_subt rrstep_rstep_pos_conv rstep_cases)

lemma rstep_args_NF_imp_rrstep:
  assumes "(s, t) \<in> rstep R"
    and "\<forall>u \<lhd> s. u \<in> NF (rstep R)"
  shows "(s, t) \<in> rrstep R"
  using assms by (metis NF_iff_no_step nrrstep_subt rstep_cases)

lemma rstep_pos_arg:
  assumes "(s, t) \<in> rstep_pos R p"
    and "i < length ss" and "ss ! i = s" 
  shows "(Fun f ss, (ctxt_of_pos_term <i> (Fun f ss))\<langle>t\<rangle>) \<in> rstep_pos R (i <# p)"
using assms
apply (cases)
apply auto
by (metis args.simps(2) ctxt_apply.simps(2) ctxt_of_pos_term.simps(2) poss_PCons_poss rstep_pos.intros subt_at.simps(2))

lemma rstep_imp_max_pos:
  assumes "(s, t) \<in> rstep R"
  shows "\<exists>u. \<exists>p\<in>poss s. (s, u) \<in> rstep_pos R p \<and> (\<forall>v \<lhd> s |_ p. v \<in> NF (rstep R))"
using assms
proof (induction s arbitrary: t)
  case (Var x)
  from Var_rstep_imp_rstep_pos_epsilon [OF this] show ?case by auto
next
  case (Fun f ss)
  show ?case
  proof (cases "\<forall>v \<lhd> Fun f ss |_ \<epsilon>. v \<in> NF (rstep R)")
    case True
    moreover with Fun.prems
      have "(Fun f ss, t) \<in> rstep_pos R \<epsilon>"
      by (auto dest: rstep_args_NF_imp_rrstep simp: rrstep_rstep_pos_conv)
    ultimately show ?thesis by auto
  next
    case False
    then obtain v where "v \<lhd> Fun f ss" and "v \<notin> NF (rstep R)" by auto
    then obtain s and w where "s \<in> set ss" and "s \<unrhd> v" and "(s, w) \<in> rstep R"
      by (auto simp: NF_def) (metis NF_iff_no_step NF_subterm supt_Fun_imp_arg_supteq)
    from Fun.IH [OF this(1, 3)] obtain u and p
      where "p \<in> poss s" and *: "(s, u) \<in> rstep_pos R p"
      and **: "\<forall>v \<lhd> s |_ p. v \<in> NF (rstep R)" by blast
    from `s \<in> set ss` obtain i
      where "i < length ss" and [simp]:"ss ! i = s" by (auto simp: in_set_conv_nth)
    with `p \<in> poss s` have "i <# p \<in> poss (Fun f ss)" by auto
    moreover with ** have "\<forall>v \<lhd> Fun f ss |_ (i <# p). v \<in> NF (rstep R)" by auto
    moreover from rstep_pos_arg [OF * `i < length ss` `ss ! i = s`]
      have "(Fun f ss, (ctxt_of_pos_term <i> (Fun f ss))\<langle>u\<rangle>) \<in> rstep_pos R (i <# p)" .
    ultimately show ?thesis by blast
  qed
qed

lemma rstep_pos_imp_rstep_pos_epsilon:
  assumes "(s, t) \<in> rstep_pos R p"
  shows "(s |_ p, t |_ p) \<in> rstep_pos R \<epsilon>"
  using assms
  apply (cases)
  apply auto
  by (metis ctxt_apply.simps(1) ctxt_of_pos_term.simps(1) empty_pos_in_poss replace_at_subt_at rstep_pos.intros)

lemma rstep_pos_imp_max_step:
  fixes s t :: "('f, 'v) term" and R :: "('f, 'v) trs"
  assumes "(s, t) \<in> rstep_pos R p"
  shows "\<exists>u. \<exists>q\<ge>p. (s, u) \<in> rstep_pos R q \<and>
    (\<forall>v \<lhd> s |_ q. s \<in> NF (rstep R))"
proof -
  from rstep_pos_imp_rstep_pos_epsilon [OF assms]
    have "(s |_ p, t |_ p) \<in> rstep R" by (auto simp: rstep_rstep_pos_conv)
  from rstep_imp_max_pos [OF this] obtain u and q
    where "q \<in> poss (s |_ p)" and "(s |_ p, u) \<in> rstep_pos R q"
    and "\<forall>v \<lhd> s |_ p |_ q. v \<in> NF (rstep R)" by blast
oops

lemma cpeaks_max_imp_PCP:
  assumes "(t, s, p, u) \<in> cpeaks R"
    and "\<forall>v \<lhd> s |_ p. v \<in> NF (rstep R)"
  shows "(t, u) \<in> PCP R"
  using assms
  unfolding cpeaks2_def PCP_def
  by (smt Pair_inject hole_pos_ctxt_of_pos_term hole_pos_in_poss mem_Collect_eq overlap_source_eq(1) overlap_source_eq(2) replace_at_subt_at subt_at_subst)

lemma cpeaks_imp_poss:
  assumes "(t, s, p, u) \<in> cpeaks R"
  shows "p \<in> poss s"
  using assms
  apply (auto simp: cpeaks2_def)
  by (metis fst_conv overlap_source_eq(1) poss_imp_subst_poss)

lemma foo:
  assumes "\<forall>v\<lhd>s \<cdot> \<sigma> |_ p. v \<in> NF (rstep R)" and "p \<in> poss s"
  shows "\<forall>v\<lhd>s |_ p. v \<in> NF (rstep R)"
  using assms
  by (auto) (metis NF_instance instance_no_supt_imp_no_supt)

lemma S3_cpeaks_max_imp_rstep_PCP:
  assumes "(t, s, p, u) \<in> S3 (cpeaks R)"
    and "\<forall>v \<lhd> s |_ p. v \<in> NF (rstep R)"
  shows "(t, u) \<in> rstep (PCP R)"
  using assms
  apply (cases)
  apply auto
  by (metis NF_instance cpeaks_imp_poss cpeaks_max_imp_PCP rstep.id rstep.subst subt_at_subst supt_stable)

lemma pos_diff_Empty2 [simp]:
  "pos_diff p \<epsilon> = p"
  by (auto simp: pos_diff_def)

lemma cpeaks_imp_nabla2:
  fixes R :: "('f, 'v :: infinite) trs"
  assumes vc: "\<forall>(l, r)\<in>R. vars_term r \<subseteq> vars_term l"
    and "(t, s, p, u) \<in> cpeaks R"
  shows "\<exists>v. nabla R s t v \<and> nabla R s v u"
proof -
  from assms obtain l\<^sub>1 r\<^sub>1 p l\<^sub>2 r\<^sub>2 and \<sigma> :: "('f, 'v) subst"
    where ol: "overlap R R (l\<^sub>1, r\<^sub>1) p (l\<^sub>2, r\<^sub>2)"
    and \<sigma>: "\<sigma> = the_mgu l\<^sub>1 (l\<^sub>2 |_ p)"
    and s: "s = l\<^sub>2 \<cdot> \<sigma>" and t: "t = replace_at l\<^sub>2 p r\<^sub>1 \<cdot> \<sigma>" and u: "u = r\<^sub>2 \<cdot> \<sigma>" 
    by (auto simp: cpeaks2_def)
  from ol have *: "(l\<^sub>1, r\<^sub>1) \<in> rstep R" "(l\<^sub>2, r\<^sub>2) \<in> rstep R"
    and **: "l\<^sub>2 \<cdot> \<sigma> = replace_at l\<^sub>2 p l\<^sub>1 \<cdot> \<sigma>"
    using overlap_source_eq [OF ol]
    by (auto dest: overlap_imp_rstep simp: \<sigma>)
  from ol have p: "p \<in> poss l\<^sub>2" by (auto simp: overlap_def dest: funposs_imp_poss)
  with ** have sp: "s |_ p = l\<^sub>1 \<cdot> \<sigma>"
    by (auto simp: s) (metis hole_pos_ctxt_of_pos_term hole_pos_subst subt_at_hole_pos)
  have peak: "(s, t) \<in> (rstep R)" "(s, u) \<in> (rstep R)"
    using * by (auto simp: s t u) (auto simp: **)
  from overlap_rstep_pos_left [OF ol]
    have stp: "(s, t) \<in> rstep_pos R p" by (auto simp: s t \<sigma>)
  from overlap_rstep_pos_right [OF ol]
    have sue: "(s, u) \<in> rstep_pos R \<epsilon>" by (auto simp: s u \<sigma>)
  show ?thesis
  proof (cases "\<forall>w \<lhd> s |_ p. w \<in> NF (rstep R)")
    txt {*@{term "(t, s, u)"} is a critical peak (at position @{term p}).*}
    case True
    with ol have "(t, u) \<in> PCP R" unfolding sp by (force simp: PCP_def s t u \<sigma>)
    with peak have "nabla R s t u" by (auto)
    with nabla_imp_nabla_refl [OF this] show ?thesis by blast
  next
    case False
    then obtain w where w: "w \<lhd> s |_ p" and "w \<notin> NF (rstep R)" by blast
    then obtain v' where "(w, v') \<in> rstep R" by blast
    from rstep_imp_max_pos [OF this] obtain w' q'
      where "q' \<in> poss w" and q'_step: "(w, w') \<in> rstep_pos R q'"
      and max: "\<forall>v \<lhd> w |_ q'. v \<in> NF (rstep R)" by blast
    from supt_imp_subt_at_nepos [OF w] obtain p'
      where "p' \<noteq> \<epsilon>" and "p' \<in> poss (s |_ p)" and "s |_ p |_ p' = w" by blast
    then have 0: "(p <#> p') \<in> poss s" and 1: "s |_ (p <#> p') = w" using p by (auto simp: s)
    def q \<equiv> "p <#> p' <#> q'"
    def v \<equiv> "(ctxt_of_pos_term (p <#> p') s)\<langle>w'\<rangle>"
    from rstep_pos_supt [OF q'_step 0 1]
      have svq: "(s, v) \<in> rstep_pos R q" by (auto simp: q_def v_def)
    then have "(s, v) \<in> rstep R" by (auto simp: rstep_rstep_pos_conv)
    have "q > p" using `p' \<noteq> \<epsilon>` by (simp add: q_def)
    from max have max: "\<forall>v \<lhd> s |_ q. v \<in> NF (rstep R)"
      using 0 and 1 by (auto simp: q_def ac_simps)
    from peak_imp_join_or_S3_cpeaks [OF vc stp svq]
      show ?thesis
    proof
      assume "(t, v) \<in> (rstep R)\<^sup>\<down>"
      then have *: "nabla R s t v" using `(s, t) \<in> rstep R` and `(s, v) \<in> rstep R` by auto
      show ?thesis
      proof (cases "(v, s, q, u) \<in> S3 (cpeaks R)")
        case True
        from S3_cpeaks_max_imp_rstep_PCP [OF this max]
          have "nabla R s v u" using `(s, v) \<in> rstep R` and `(s, u) \<in> rstep R` by auto
        with * show ?thesis by blast
      next
        case False
        with peak_imp_join_or_S3_cpeaks [OF vc svq sue]
          have "(v, u) \<in> (rstep R)\<^sup>\<down>" apply auto
          by (metis `p < q` epsilon_least eq_iff less_simps(3))
        then have "nabla R s v u" using `(s, v) \<in> rstep R` and `(s, u) \<in> rstep R` by auto
        with * show ?thesis by blast
      qed
    next
      assume "\<exists>va w.
       q \<le> p \<and>
       t = (ctxt_of_pos_term q s)\<langle>va\<rangle> \<and>
       v = (ctxt_of_pos_term q s)\<langle>w\<rangle> \<and>
       (va, s |_ q, pos_diff p q, w) \<in> S3 (cpeaks R) \<or>
       p \<le> q \<and>
       t = (ctxt_of_pos_term p s)\<langle>va\<rangle> \<and>
       v = (ctxt_of_pos_term p s)\<langle>w\<rangle> \<and>
       (w, s |_ p, pos_diff q p, va) \<in> S3 (cpeaks R)"
      with `q > p` obtain x y
        where t: "t = (ctxt_of_pos_term p s)\<langle>x\<rangle>"
        and v: "v = (ctxt_of_pos_term p s)\<langle>y\<rangle>"
        and *: "(y, s |_ p, pos_diff q p, x) \<in> S3 (cpeaks R)" by auto
      from S3_cpeaks_max_imp_rstep_PCP [OF *] and max
        have "(y, x) \<in> rstep (PCP R)"
        by (metis "0" `p < q` dual_order.order_iff_strict poss_append_poss prefix_pos_diff subt_at_append)
      then have "(t, v) \<in> (rstep (PCP R))\<^sup>\<leftrightarrow>" by (auto simp: t v)
      then have *: "nabla R s t v" using `(s, t) \<in> rstep R` and `(s, v) \<in> rstep R` by auto
      show ?thesis
      proof (cases "(v, s, q, u) \<in> S3 (cpeaks R)")
        case True
        from S3_cpeaks_max_imp_rstep_PCP [OF this max]
          have "nabla R s v u" using `(s, v) \<in> rstep R` and `(s, u) \<in> rstep R` by auto
        with * show ?thesis by blast
      next
        case False
        with peak_imp_join_or_S3_cpeaks [OF vc svq sue]
          have "(v, u) \<in> (rstep R)\<^sup>\<down>" apply auto
          by (metis `p < q` epsilon_least eq_iff less_simps(3))
        then have "nabla R s v u" using `(s, v) \<in> rstep R` and `(s, u) \<in> rstep R` by auto
        with * show ?thesis by blast
      qed
    qed
  qed
qed

lemma CS3_cpeaks_imp_nabla2:
  assumes vc: "\<forall>(l, r)\<in>R. vars_term r \<subseteq> vars_term l"
    and *: "(t, s, p, u) \<in> CS3 (cpeaks R)"
  shows "\<exists>v. nabla R s t v \<and> nabla R s v u"
proof -
  from * obtain C t' s' p' u' \<sigma> where cp: "(t', s', p', u') \<in> cpeaks R"
    and t: "t = C\<langle>t' \<cdot> \<sigma>\<rangle>" and s: "s = C\<langle>s' \<cdot> \<sigma>\<rangle>" and u: "u = C\<langle>u' \<cdot> \<sigma>\<rangle>"
    and p: "p = hole_pos C <#> p'"
    by (cases)
  from cpeaks_imp_nabla2 [OF vc cp] obtain v
    where "nabla R s' t' v" and "nabla R s' v u'" by blast
  then have "nabla R s t (C\<langle>v \<cdot> \<sigma>\<rangle>)" and "nabla R s (C\<langle>v \<cdot> \<sigma>\<rangle>) u"
    unfolding s t u by (auto intro: nabla_subst nabla_ctxt)
  then show ?thesis by blast
qed

lemma nabla_commute:
  "nabla R s t u = nabla R s u t"
  by (auto simp: nabla_def)

(*Lemma 5.1.15*)
lemma peak_imp_nabla2:
  assumes vc: "\<forall>(l, r)\<in>R. vars_term r \<subseteq> vars_term l"
    and peak: "(s, t) \<in> rstep R" "(s, u) \<in> rstep R"
  shows "\<exists>v. nabla R s t v \<and> nabla R s v u"
proof -
  { assume "(t, u) \<in> (rstep R)\<^sup>\<down>"
    with peak have "nabla R s t u" by (auto simp: nabla_def)
    with nabla_imp_nabla_refl [OF this] have ?thesis by blast }
  moreover
  { fix p assume "(t, s, p, u) \<in> CS3 (cpeaks R)"
    from CS3_cpeaks_imp_nabla2 [OF vc this] have ?thesis . } 
  moreover
  { fix p assume "(u, s, p, t) \<in> CS3 (cpeaks R)"
    from CS3_cpeaks_imp_nabla2 [OF vc this] have ?thesis
      by (auto simp: nabla_commute) }
  ultimately show ?thesis
    using peak_imp_join_or_CS3_cpeaks [OF vc peak] by blast
qed

end

