theory Prime_Product
imports 
  Sqrt_Babylonian
  NthRoot_Impl
  "~~/src/HOL/Number_Theory/UniqueFactorization"
begin

(* missing library *)

lemma multiplicity_prime: assumes p: "prime (i :: nat)" and ji: "j \<noteq> i"
  shows "multiplicity j i = 0"
proof (rule ccontr)
  assume "\<not> ?thesis"
  hence "multiplicity j i > 0" by auto
  hence j: "j \<in> prime_factors i"
    by (metis less_not_refl multiplicity_not_factor_nat)
  hence d: "j dvd i" 
    by (metis p prime_factors_altdef2_nat prime_gt_0_nat)
  then obtain k where i: "i = j * k" ..
  from j have "j \<ge> 2" 
    by (metis prime_factors_prime_nat prime_ge_2_nat)
  hence j1: "j \<noteq> 1" by auto
  from i have "j dvd i" by auto
  with j1 ji p[unfolded prime_nat_def] show False by auto
qed
 
(* end missing library *)

definition prime_product where "prime_product (n :: nat) = (\<forall> p. multiplicity p n \<le> 1)"

function prime_product_factor_main :: "nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> nat \<times> nat" where
  "prime_product_factor_main factor_sq factor_pr limit n i = 
    (if i \<le> limit \<and> i \<ge> 2 then
       (if i dvd n 
           then (let n' = n div i in 
             (if i dvd n' then
                let n'' = n' div i in 
                  prime_product_factor_main (factor_sq * i) factor_pr (nat (root_nat_floor 3 n'')) n'' i
                else
                  (case sqrt_nat n' of 
                    Cons sn _ \<Rightarrow> (factor_sq * sn, factor_pr * i)
                  | [] \<Rightarrow> prime_product_factor_main factor_sq (factor_pr * i) (nat (root_nat_floor 3 n')) n' (Suc i)
                  )
             )              
           )
         else 
           prime_product_factor_main factor_sq factor_pr limit n (Suc i))
       else
         (factor_sq, factor_pr * n))" by pat_completeness auto

termination
proof -
  let ?m1 = "\<lambda> (factor_sq :: nat,factor_pr :: nat,limit :: nat,n :: nat,i :: nat). n"
  let ?m2 = "\<lambda> (factor_sq,factor_pr,limit,n,i). (Suc limit - i)"
  {
    fix i
    have "2 \<le> i \<Longrightarrow> Suc 0 < i * i" using one_less_mult[of i i] by auto
  } note * = this
  show ?thesis
    by (rule, rule wf_measures[of "[?m1,?m2]"], auto split: if_splits simp: * dvd_eq_mod_eq_0)
qed
  
lemma prime_product_factor_main: assumes "\<not> (\<exists> s. s * s = n)"
  and "limit = nat (root_nat_floor 3 n)"
  and "m = factor_sq * factor_sq * factor_pr * n"
  and "prime_product_factor_main factor_sq factor_pr limit n i = (sq, p)"
  and "i \<ge> 2"
  and "\<And> j. j \<ge> 2 \<Longrightarrow> j < i \<Longrightarrow> \<not> j dvd n"
  and "\<And> j. j < i \<Longrightarrow> multiplicity j factor_pr \<le> 1"
  and "\<And> j. j \<ge> i \<Longrightarrow> multiplicity j factor_pr = 0"
  and "factor_pr > 0"
  shows "m = sq * sq * p \<and> prime_product p"
  using assms
proof (induct factor_sq factor_pr limit n i rule: prime_product_factor_main.induct)
  case (1 factor_sq factor_pr limit n i)
  note IH = 1(1-3)
  note prems = 1(4-)
  note simp = prems(4)[unfolded prime_product_factor_main.simps[of factor_sq factor_pr limit n i]]
  show ?case
  proof (cases "i \<le> limit")
    case True note i = this
    with prems(5) have cond: "i \<le> limit \<and> i \<ge> 2" and *: "(i \<le> limit \<and> i \<ge> 2) = True" by blast+
    note IH = IH[OF cond]
    note simp = simp[unfolded * if_True]
    show ?thesis
    proof (cases "i dvd n")
      case False
      hence *: "(i dvd n) = False" by simp
      note simp = simp[unfolded * if_False]      
      note IH = IH(3)[OF False prems(1-3) simp]
      show ?thesis
      proof (rule IH)
        fix j
        assume 2: "2 \<le> j" and j: "j < Suc i"
        from prems(6)[OF 2] j False
        show "\<not> j dvd n" by (cases "j = i", auto)
      next
        fix j
        assume j: "j < Suc i"
        with prems(7-8)[of j]  
        show "multiplicity j factor_pr \<le> 1" by (cases "j = i", auto)
      qed (insert prems(8-9) cond, auto)
    next
      case True note mod = this
      hence "(i dvd n) = True" by simp
      note simp = simp[unfolded this if_True Let_def]
      note IH = IH(1,2)[OF True refl]
      show ?thesis
      proof (cases "i dvd n div i")
        case True
        hence *: "(i dvd n div i) = True" by auto
        def n' \<equiv> "n div i div i"
        from mod True have n: "n = n' * i * i" by (auto simp: n'_def dvd_eq_mod_eq_0)
        note simp = simp[unfolded * if_True split]
        note IH = IH(1)[OF True refl _ refl _ simp prems(5) _ prems(7-9)]
        show ?thesis
        proof (rule IH)
          show "m = factor_sq * i * (factor_sq * i) * factor_pr * (n div i div i)"
            unfolding prems(3) n'_def[symmetric]
            unfolding n by (auto simp: field_simps)
        next
          fix j
          assume "2 \<le> j" "j < i"
          from prems(6)[OF this] have "\<not> j dvd n" by auto
          thus "\<not> j dvd n div i div i" 
            by (metis dvd_mult n n'_def nat_mult_assoc nat_mult_commute)
        next
          show "\<not> (\<exists> s. s * s = n div i div i)"
          proof
            assume "\<exists> s. s * s = n div i div i"
            then obtain s where "s * s = n div i div i" by auto
            hence "(s * i) * (s * i) = n" unfolding n by auto
            with prems(1) show False by blast
          qed
        qed
      next
        case False        
        def n' \<equiv> "n div i"
        from mod True have n: "n = n' * i" by (auto simp: n'_def dvd_eq_mod_eq_0)
        have prime: "prime i" 
          unfolding prime_nat_def
        proof (intro conjI allI impI)
          fix m
          assume m: "m dvd i"
          hence "m dvd n" unfolding n by auto
          with prems(6)[of m] have choice: "m \<le> 1 \<or> m \<ge> i" by arith
          from m prems(5) have "m > 0"
            by (metis dvd.order.not_eq_order_implies_strict dvd_0_right neq_zero_eq_gt_zero_nat not_numeral_le_zero)
          with choice have choice: "m = 1 \<or> m \<ge> i" by arith
          from m prems(5) have "m \<le> i" 
            by (metis False div_by_0 dvd.dual_order.refl dvd_imp_le gr0I)
          with choice
          show "m = 1 \<or> m = i" by auto        
        qed (insert prems(5), auto)
        from False have "(i dvd n div i) = False" by auto
        note simp = simp[unfolded this if_False]
        note IH = IH(2)[OF False _ _ refl]
        from prime have "i > 0" by auto
        note mult = multiplicity_product_nat[OF prems(9) this]
        note mult_i = multiplicity_prime_nat[OF prime] multiplicity_prime[OF prime]
        show ?thesis
        proof (cases "sqrt_nat (n div i)")
          case (Cons s)
          note simp = simp[unfolded Cons list.simps]
          hence sq: "sq = factor_sq * s" and p: "p = factor_pr * i" by auto          
          from arg_cong[OF Cons, of set] have s: "s * s = n div i" by auto
          have pp: "prime_product (factor_pr * i)" 
            unfolding prime_product_def
          proof
            fix m
            show "multiplicity m (factor_pr * i) \<le> 1"
              unfolding mult using prems(7)[of m] prems(8)[of m] mult_i(1) mult_i(2)[of m] by fastforce
          qed
          show ?thesis unfolding sq p prems(3) n unfolding n'_def s[symmetric]
            using pp by auto
        next
          case Nil
          note simp = simp[unfolded Nil list.simps]
          from arg_cong[OF Nil, of set] have "\<not> (\<exists> x. x * x = n div i)" by simp
          note IH = IH[OF Nil this  _ simp] 
          show ?thesis
          proof (rule IH)
            show "m = factor_sq * factor_sq * (factor_pr * i) * (n div i)" 
              unfolding prems(3) n by auto
          next
            fix j
            assume *: "2 \<le> j" "j < Suc i"
            show "\<not> j dvd n div i"
            proof
              assume j: "j dvd n div i"
              with False have "j \<noteq> i" by auto
              with * have "2 \<le> j" "j < i" by auto
              from prems(6)[OF this] j
              show False unfolding n
                by (metis dvd_mult n n'_def nat_mult_commute)
            qed
          next
            fix j
            assume "Suc i \<le> j"
            hence ij: "i \<le> j" and j: "j \<noteq> i" by auto
            have 0: "multiplicity j i = 0" using prime j by (rule multiplicity_prime)
            show "multiplicity j (factor_pr * i) = 0" unfolding mult prems(8)[OF ij] 0 by simp
          next
            fix j
            assume "j < Suc i"
            hence "j < i \<or> j = i" by auto
            thus "multiplicity j (factor_pr * i) \<le> 1"
            proof 
              assume "j = i"
              with prems(8)[of i] prime show ?thesis unfolding mult
                by (auto)
            next
              assume ji: "j < i"
              hence "j \<noteq> i" by auto
              from prems(7)[OF ji] multiplicity_prime[OF prime this]
              show ?thesis unfolding mult by auto
            qed
          qed (insert prems(5,9), auto)
        qed
      qed
    qed
  next
    case False
    hence "(i \<le> limit \<and> i \<ge> 2) = False" by auto
    note simp = simp[unfolded this if_False]
    hence sq: "sq = factor_sq" and p: "p = factor_pr * n" by auto
    show ?thesis
    proof
      show "m = sq * sq * p" unfolding sq p prems(3) by simp
      show "prime_product p" unfolding prime_product_def
      proof
        fix m
        from prems(1) have n1: "n > 1" by (cases n, auto, case_tac nat, auto)
        hence n0: "n > 0" by auto
        have "i > limit" using False by auto
        from this[unfolded prems(2)] have less: "int i \<ge> root_nat_floor 3 n + 1" by auto
        have "int n < (root_nat_floor 3 n + 1) ^ 3" by (rule root_nat_floor_upper, auto)
        also have "\<dots> \<le> int i ^ 3" by (rule power_mono[OF less, of 3], auto)
        finally have n_i3: "n < i ^ 3"
          by (metis zless_int zpower_int)
        {
          fix m
          assume m: "multiplicity m n > 0"
          hence mp: "m \<in> prime_factors n"
            by (metis less_not_refl multiplicity_not_factor_nat)
          hence md: "m dvd n" 
            by (metis n0 prime_factors_altdef2_nat)
          then obtain k where n: "n = m * k" ..
          from mp have pm: "prime m" by auto
          hence m2: "m \<ge> 2" and m0: "m > 0" by auto
          from prems(6)[OF m2] md have mi: "m \<ge> i" by force
          {
            assume "multiplicity m n \<noteq> 1"
            with m have "\<exists> k. multiplicity m n = 2 + k" by presburger
            then obtain j where mult: "multiplicity m n = 2 + j" ..
            from n0 n have k: "k > 0" by auto
            from mult[unfolded n multiplicity_product_nat[OF m0 k]] pm 
            have "multiplicity m k > 0" by auto
            hence mp: "m \<in> prime_factors k"
              by (metis less_not_refl multiplicity_not_factor_nat)
            hence md: "m dvd k" 
              by (metis k prime_factors_altdef2_nat)            
            then obtain l where kml: "k = m * l" ..
            note n = n[unfolded kml]
            from n have "l dvd n" by auto
            with prems(6)[of l] have "l \<le> 1 \<or> l \<ge> i" by arith
            with n n0 have l: "l = 1 \<or> l \<ge> i" by auto
            from n prems(1) have "l \<noteq> 1" by auto
            with l have l: "l \<ge> i" by auto
            from mult_le_mono[OF mult_le_mono[OF mi mi] l]
            have "n \<ge> i^3" unfolding n by (auto simp: power3_eq_cube)
            with n_i3 have False by auto
          }
          with mi m
          have "multiplicity m n = 1 \<and> m \<ge> i" by auto
        } note n = this
        have "multiplicity m p = multiplicity m factor_pr + multiplicity m n"
          unfolding p
          by (rule multiplicity_product_nat[OF prems(9)], insert prems(1), cases n, auto)
        also have "\<dots> \<le> 1"
        proof (cases "m < i")
          case True
          from prems(7)[of m] n[of m] True show ?thesis by force
        next
          case False
          hence "m \<ge> i" by auto
          from prems(8)[OF this] n[of m] show ?thesis by force
        qed
        finally show "multiplicity m p \<le> 1" .
      qed
    qed
  qed
qed


definition prime_product_factor :: "nat \<Rightarrow> nat \<times> nat" where
  "prime_product_factor n = (case sqrt_nat n of 
     (Cons s _) \<Rightarrow> (s,1)
   | [] \<Rightarrow> prime_product_factor_main 1 1 (nat (root_nat_floor 3 n)) n 2)"

lemma prime_product_one[simp, intro]: "prime_product 1"
  unfolding prime_product_def multiplicity_one_nat by auto

lemma prime_product_factor: assumes pf: "prime_product_factor n = (sq,p)"
  shows "n = sq * sq * p \<and> prime_product p"
proof (cases "sqrt_nat n")
  case (Cons s) 
  from pf[unfolded prime_product_factor_def Cons] arg_cong[OF Cons, of set]
    prime_product_one
  show ?thesis by auto
next
  case Nil
  from arg_cong[OF Nil, of set] have nsq: "\<not> (\<exists>s. s * s = n)" by auto
  show ?thesis
    by (rule prime_product_factor_main[OF nsq refl, of _ 1 1 2], unfold multiplicity_one_nat,
    insert pf[unfolded prime_product_factor_def Nil], auto)
qed

end
